"""
implementation of log sv model calibration to time series
need package option_chain_analytics with BTC and ETH data
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import qis
from typing import Dict, Tuple, Any, Optional
from enum import Enum

# analytics
from stochvolmodels.data.fetch_option_chain import (generate_vol_chain_np,
                                                    load_option_chain,
                                                    sample_option_chain_at_times,
                                                    load_price_data)
from stochvolmodels import (OptionChain, LogSvParams, LogSVPricer, ConstraintsType, LogsvModelCalibrationType)

# chain data
from option_chain_analytics import OptionsDataDFs, create_chain_from_from_options_dfs
from option_chain_analytics.ts_loaders import ts_data_loader_wrapper


def calibrate_logsv_model_with_fixed_kappas(option_chain: OptionChain,
                                            kappa1: float = 2.21,
                                            kappa2: float = 2.18
                                            ) -> LogSvParams:
    """
    calibrate logsv model for option chain with keeeping fixed kappa1 and kappa2z
    """
    logsv_pricer = LogSVPricer()
    # use atm vols and skews to set initial values and bounds
    atm_vols = option_chain.get_chain_atm_vols()
    skews = option_chain.get_chain_skews(delta=0.4)
    params0 = LogSvParams(sigma0=atm_vols[0], theta=atm_vols[-1], kappa1=kappa1, kappa2=kappa2, beta=-2.0*np.nanmean(skews), volvol=1.0)
    calibrated_params = logsv_pricer.calibrate_model_params_to_chain(
        option_chain=option_chain,
        params0=params0,
        params_min=LogSvParams(sigma0=0.9*atm_vols[0], theta=0.9*atm_vols[-1], kappa1=0.25, kappa2=0.25, beta=-3.0, volvol=0.9),
        params_max=LogSvParams(sigma0=1.1*atm_vols[0], theta=1.1*atm_vols[-1], kappa1=10.0, kappa2=10.0, beta=3.0, volvol=6.0),
        constraints_type=ConstraintsType.MMA_MARTINGALE,
        model_calibration_type=LogsvModelCalibrationType.PARAMS4)

    return calibrated_params


def plot_calibration_report(option_chain: OptionChain,
                            params: LogSvParams,
                            value_time: pd.Timestamp,
                            is_detailed: bool = True
                            ) -> Tuple[Dict[str, Any], Dict[str, plt.Figure]]:
    logsv_pricer = LogSVPricer()
    fig1 = logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain, params=params)
    if is_detailed:
        qis.set_suptitle(fig=fig1, title=f"Model fit @ {value_time}: {params.to_str()}")
    fig2 = logsv_pricer.plot_model_ivols_vs_mc(option_chain=option_chain, params=params, nb_path=400000)
    if is_detailed:
        qis.set_suptitle(fig=fig2, title=f"Model vs MC @ {value_time}: {params.to_str()}")
    figs_dict = dict()
    figs_dict[f"fig1_{value_time}"] = fig1
    figs_dict[f"fig2_{value_time}"] = fig2

    # compute mse
    vol_scaler = logsv_pricer.set_vol_scaler(option_chain=option_chain)
    model_ivols = logsv_pricer.compute_model_ivols_for_chain(option_chain=option_chain, params=params, vol_scaler=vol_scaler)

    output_dict = params.to_dict()
    # add fitted sliced ids
    for idx, slice_id in enumerate(option_chain.ids):
        output_dict[f"slice-{idx+1} id"] = slice_id

    # add mses
    mse2s = []
    vol_spreads = []
    for idx, ttm in enumerate(option_chain.ttms):
        midvols = 0.5 * (option_chain.bid_ivs[idx] + option_chain.ask_ivs[idx])
        vol_spread = 0.5*(option_chain.ask_ivs[idx] - option_chain.bid_ivs[idx])
        mse2 = np.sqrt(np.nanmean(np.power(model_ivols[idx] - midvols, 2)))
        output_dict[f"slice-{idx+1} mse"] = mse2
        mse2s.append(mse2)
        vol_spreads.append(np.nanmean(vol_spread))
    output_dict['avg mse'] = np.nanmean(np.array(mse2s))
    output_dict['avg vol-spread'] = np.nanmean(np.array(vol_spreads))

    # add atm vols
    atm_vols = option_chain.get_chain_atm_vols()
    output_dict.update({f"atm_vol-{idx+1}": vol for idx, vol in enumerate(atm_vols)})

    # add skews
    skews = option_chain.get_chain_skews(delta=0.4)
    output_dict.update({f"skew-{idx+1}": skew for idx, skew in enumerate(skews)})

    return output_dict, figs_dict


def run_calibration_time_series(options_data_dfs: OptionsDataDFs,
                                time_period: qis.TimePeriod,
                                freq: str = 'W-FRI',
                                hour_offset: int = 10,
                                days_map: Dict[str, int] = {'1w': 7, '2w': 14, '1m': 21},
                                delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.1, 0.1),
                                kappa1: float = 2.21,
                                kappa2: float = 2.18
                                ) -> Tuple[pd.DataFrame, Dict[str, plt.Figure]]:
    """
    generate chains at given freqs and report calibration stats
    """

    option_chains = sample_option_chain_at_times(options_data_dfs=options_data_dfs,
                                                 time_period=time_period,
                                                 freq=freq,
                                                 hour_offset=hour_offset,
                                                 days_map=days_map,
                                                 delta_bounds=delta_bounds)
    output_dicts = dict()
    figs_dicts = dict()
    for value_time, option_chain in option_chains.items():
        print(f"{value_time}")
        params = calibrate_logsv_model_with_fixed_kappas(option_chain=option_chain, kappa1=kappa1, kappa2=kappa2)
        output_dict, figs_dict = plot_calibration_report(option_chain=option_chain,
                                                         params=params,
                                                         value_time=value_time)
        print(output_dict)
        output_dicts[value_time] = pd.Series(output_dict)
        figs_dicts.update(figs_dict)
        plt.close('all')
    output_df = pd.DataFrame.from_dict(output_dicts, orient='index')
    return output_df, figs_dicts


def report_calibration_timeseries(df: pd.DataFrame,
                                  fontsize: int = 14
                                  ) -> plt.Figure:

    df = df.drop(['13/03/2020  10:00:00', '15/01/2021  10:00:00'], axis=0)

    vols = df[['sigma0', 'theta']].rename({'sigma0': r'$\sigma_0$', 'theta': r'$\theta$'}, axis=1)
    beta = df['beta'].rename(r'$\beta$')
    volvol = df['volvol'].rename(r'$\varepsilon$')
    avg_mse = df[['avg mse', 'avg vol-spread']].rename({'avg mse': 'Avg model MSE', 'avg vol-spread': 'Avg Bid/Ask Spread'}, axis=1)
    avg_mse['Avg Bid/Ask Spread'] *= 2.0

    kwargs = dict(framealpha=0.9, fontsize=fontsize)
    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(2, 2, figsize=(18, 10), tight_layout=True)
        qis.plot_time_series(df=avg_mse,
                             var_format='{:.2%}',
                             title='(A) Average mean-squared error and bid-ask volatility spread',
                             ax=axs[0][0],
                             **kwargs)
        qis.plot_time_series(df=vols,
                             var_format='{:.2%}',
                             title='(B) Initial and mean volatilities',
                             ax=axs[1][0],
                             **kwargs)
        qis.plot_time_series(df=beta,
                             ax=axs[0][1],
                             title='(C) Volatility beta',
                             **kwargs)
        qis.plot_time_series(df=volvol,
                             ax=axs[1][1],
                             title='(D) Volatility-of-volatility',
                             **kwargs)
    return fig


class UnitTests(Enum):
    FIT_LOGSV_MODEL = 1
    REPORT_FITTED_MODEL = 2
    RUN_CALIBRATION_TIMESERIES = 3
    REPORT_CALIBRATION_TIMESERIES = 4


def run_unit_test(unit_test: UnitTests):

    pd.set_option('display.max_columns', 1000)

    ticker = 'BTC'  # BTC, ETH
    value_time = pd.Timestamp('2019-06-28 10:00:00+00:00')
    value_time = pd.Timestamp('2021-01-29 10:00:00+00:00')
    value_time = pd.Timestamp('2022-12-30 10:00:00+00:00')
    value_time = pd.Timestamp('2023-06-30 10:00:00+00:00')

    # chain data here
    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker))
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)

    option_chain = generate_vol_chain_np(chain=chain,
                                         value_time=value_time,
                                         days_map={'1w': 7, '2w': 14, '1m': 21},
                                         # days_map={'2w': 14, '1m': 21},
                                         delta_bounds=(-0.1, 0.1),
                                         is_filtered=True)

    btc_kappas = dict(kappa1=2.21, kappa2=2.18)
    eth_kappas = dict(kappa1=1.96, kappa2=1.98)
    kappas = btc_kappas

    if unit_test == UnitTests.FIT_LOGSV_MODEL:
        option_chain.print()
        calibrated_params = calibrate_logsv_model_with_fixed_kappas(option_chain=option_chain, **kappas)
        print(calibrated_params)
        plot_calibration_report(option_chain=option_chain,
                                params=calibrated_params,
                                value_time=value_time)

    elif unit_test == UnitTests.REPORT_FITTED_MODEL:
        option_chain.print()
        params = LogSvParams(sigma0=0.4083, theta=0.3789, kappa1=2.21, kappa2=2.18, beta=0.5010, volvol=3.0633)
        output_dict, figs_dict = plot_calibration_report(option_chain=option_chain,
                                                         params=params,
                                                         value_time=value_time,
                                                         is_detailed=False)
        print(output_dict)

        qis.save_figs_to_pdf(figs=figs_dict, file_name='btc_calibration', local_path=f"C://Users//artur//OneDrive//analytics//outputs//")

    elif unit_test == UnitTests.RUN_CALIBRATION_TIMESERIES:
        time_period = qis.TimePeriod('01Jan2023', '31Jan2023', tz='UTC')
        time_period = qis.TimePeriod('2019-03-30 00:00:00+00:00', '2023-09-01 00:00:00+00:00', tz='UTC')

        output_df, figs_dict = run_calibration_time_series(options_data_dfs=options_data_dfs, time_period=time_period,
                                                           freq='W-FRI',
                                                           **kappas)
        print(output_df)

        file_name = 'eth_calibration_w_fri'
        qis.save_df_to_excel(data=output_df, file_name=file_name, local_path=f"C://Users//artur//OneDrive//analytics//outputs//",
                             add_current_date=True)
        qis.save_figs_to_pdf(figs=figs_dict, file_name=file_name, local_path=f"C://Users//artur//OneDrive//analytics//outputs//")

    elif unit_test == UnitTests.REPORT_CALIBRATION_TIMESERIES:

        df = qis.load_df_from_excel(file_name='btc_calibration_w_fri_20231209_1504', local_path=f"C://Users//artur//OneDrive//analytics//resources//")
        # df = qis.load_df_from_excel(file_name='eth_calibration_w_fri_20231210_1124', local_path=f"C://Users//artur//OneDrive//analytics//resources//")
        fig = report_calibration_timeseries(df=df)
        qis.save_fig(fig=fig, file_name='btc_calibration_w', local_path=f"C://Users//artur//OneDrive//My Papers//Working Papers//LogNormal Stochastic Volatility. London. Oct 2013//figures//")

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.REPORT_CALIBRATION_TIMESERIES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
example of fitting GMM
see StochVolModels/examples/run_gmm_fit.py
"""
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from stochvolmodels import (get_btc_test_chain_data,
                            get_spy_test_chain_data,
                            OptionChain, GmmPricer,
                            plot_gmm_pdfs)

# get test option chain data
# option_chain = get_btc_test_chain_data()
option_chain = get_spy_test_chain_data()

# run GMM fit
gmm_pricer = GmmPricer()
fit_params = gmm_pricer.calibrate_model_params_to_chain(option_chain=option_chain, n_mixtures=4)

# illustrate fitted parameters and model fit to market bid-ask
# plot two ids
ids = ['2m', '6m']
n = len(ids)
with sns.axes_style('darkgrid'):
    fig, axs = plt.subplots(n, 2, figsize=(14, 12), tight_layout=True)
    # axs = qis.to_flat_list(axs)
current_ax = 0

for key, params in fit_params.items():
    print(f"{key}: {params}")
    if key in ids:
        option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[key])
        # gmm_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain0, params=params, axs=[axs[idx]])
        plot_gmm_pdfs(params=params, option_chain0=option_chain0, axs=axs[current_ax, :])
        qis.set_title(ax=axs[current_ax, 0], title=f"{key}-slice: (A) State PDF and Aggregate Risk-Neutral PDF")
        qis.set_title(ax=axs[current_ax, 1], title=f"{key}-slice: Model to Market Bid/Ask vols")
        current_ax += 1

qis.set_suptitle(fig, title='Fit of 4-state GMM to SPY implied vols @ 15_Jul_2022_10_23_09')
plt.show()



"""
figure 5 in the logsv_model_wtih_quadratic_drift
plot steady state pdfs
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.special as sps
import matplotlib.ticker as mticker
from numba import njit
from enum import Enum
from typing import Dict

from stochvolmodels.pricers.logsv_pricer import LogSvParams
import stochvolmodels.utils.plots as plot

VOLVOL = 1.5

TEST_PARAMS = {'$(\kappa_{1}=4, \kappa_{2}=0); \sigma_{0}=2.0$': LogSvParams(sigma0=2.0, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
               '$(\kappa_{1}=4, \kappa_{2}=4); \sigma_{0}=2.0$': LogSvParams(sigma0=2.0, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
               '$(\kappa_{1}=4, \kappa_{2}=8); \sigma_{0}=2.0$': LogSvParams(sigma0=2.0, theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}

SS_PDF_PARAMS = {'$(\kappa_{1}=4, \kappa_{2}=0)$': LogSvParams(theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
                 '$(\kappa_{1}=4, \kappa_{2}=4)$': LogSvParams(theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
                 '$(\kappa_{1}=4, \kappa_{2}=8)$': LogSvParams(theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}

SS_PARAMS = {'$\kappa_{1}=1$': LogSvParams(theta=1.0, kappa1=1.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
             '$\kappa_{1}=4$': LogSvParams(theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
             '$\kappa_{1}=8$': LogSvParams(theta=1.0, kappa1=8.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}


@njit(cache=False, fastmath=True)
def integral_x_over_sigma(x_grid: np.ndarray,
                          sigma: np.ndarray,
                          g_sigma: np.ndarray
                          ) -> np.ndarray:
    pdf = np.zeros_like(x_grid)
    sigma_inv = np.reciprocal(sigma)
    den = (1.0 / np.sqrt(2.0*np.pi)) * sigma_inv
    for idx, x_ in enumerate(x_grid):
        pdf[idx] = np.nansum(den * np.exp(-0.5*(x_*x_)*sigma_inv) * g_sigma)
    return pdf


def vol_moment(params: LogSvParams, r: int = 1):
    nu = 2.0 * (params.kappa2 * params.theta - params.kappa1) / params.vartheta2 - 1.0
    q = 2.0 * params.kappa1 * params.theta / params.vartheta2
    b = 2.0 * params.kappa2 / params.vartheta2
    y = np.power(b/q, r/2.0) * sps.kv(nu+r, 2.0*np.sqrt(q*b)) / sps.kv(nu, 2.0*np.sqrt(q*b))
    return y


def vol_skeweness(params: LogSvParams):
    m3_r = vol_moment(params=params, r=3)
    m2_r = vol_moment(params=params, r=2)
    m1_r = vol_moment(params=params, r=1)
    m2 = m2_r - m1_r * m1_r
    # y = (m3_r-3*m1_r*m2_r-m1_r*m1_r*m1_r)/np.power(m2_r, 1.5)
    y = (m3_r-3*m1_r*m2-m1_r*m1_r*m1_r)/np.power(m2, 1.5)
    # y = (m3_r - 3.0 * m1_r * m2_r - 2.0*m1_r * m1_r * m1_r) / np.power(m2, 1.5)
    # print(f"m1_r={m1_r}, m2_r={m2_r}, m2={m2}, m3_r={m3_r}, skew={y}")
    return y


def steady_state(sigma: np.ndarray,
                 params: LogSvParams
                 ) -> np.ndarray:
    nu = 2.0*(params.kappa2*params.theta-params.kappa1)/params.vartheta2 - 1.0
    q = 2.0*params.kappa1*params.theta/params.vartheta2
    b = 2.0*params.kappa2/params.vartheta2
    if params.kappa1 >= 1e-6:
        if params.kappa2 >= 1e-6:
            c = np.power(b/q, nu/2.0) / (2.0*sps.kv(nu, 2.0*np.sqrt(q*b)))
        else:
            c = np.power(q, -nu) / sps.gamma(-nu)
    else:
        raise NotImplementedError(f"kappa1 = 0 is not implemented")
    g = c*np.power(sigma, nu-1.0)*np.exp(-q*np.reciprocal(sigma)-b*sigma)
    return g


def plot_steady_state(params_dict: Dict[str, LogSvParams] = SS_PDF_PARAMS,
                      title: str = None,
                      ax: plt.Subplot = None
                      ) -> None:

    sigma = np.linspace(1e-4, 4.0, 1000)
    qs = []
    for key, params in params_dict.items():
        qs.append(pd.Series(steady_state(sigma=sigma, params=params), index=sigma, name=key))
    ss_pdf = pd.concat(qs, axis=1)

    sns.lineplot(data=ss_pdf, dashes=False, ax=ax)
    ax.set_xlim(left=0.0)
    ax.set_ylim(bottom=0.0)
    ax.set_title(title, color='darkblue')
    ax.set_xlabel('$\sigma$', fontsize=12)
    yvar_format = '{:.2f}'
    ax.xaxis.set_major_formatter(mticker.FuncFormatter(lambda z, _: yvar_format.format(z)))


def plot_steady_state_x(params_dict: Dict[str, LogSvParams] = SS_PDF_PARAMS,
                        title: str = None,
                        ax: plt.Subplot = None
                        ) -> None:

    sigma = np.linspace(1e-4, 5.0, 1000)
    x = np.linspace(-5.0, 5.0, 200)
    qs = []
    for key, params in params_dict.items():
        g_sigma = steady_state(sigma=sigma, params=params)
        x_pdf = integral_x_over_sigma(x_grid=x, sigma=sigma, g_sigma=g_sigma)
        qs.append(pd.Series(x_pdf, index=x, name=key))

    ss_pdf = pd.concat(qs, axis=1)

    if ax is None:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(10, 4.0), tight_layout=True)
    sns.lineplot(data=ss_pdf, ax=ax)
    #ax.set_xlim(left=0.0)
    ax.set_ylim(bottom=0.0)
    ax.set_title(title, color='darkblue')


def plot_vol_skew(params_dict=SS_PARAMS,
                  title: str = f'Skeweness of volatility as function of $\kappa_{2}$',
                  ax: plt.Subplot = None
                  ) -> None:

    kappa2s = np.linspace(0.5, 10.0, 100)
    qs = []

    def skewness(params: LogSvParams) -> np.ndarray:
        skew = np.zeros_like(kappa2s)
        for idx, kappa2 in enumerate(kappa2s):
            params.kappa2 = kappa2
            m3_r = vol_moment(params=params, r=3)
            m2_r = vol_moment(params=params, r=2)
            m1_r = vol_moment(params=params, r=1)
            m2 = m2_r - m1_r * m1_r
            # y = (m3_r-3*m1_r*m2_r-m1_r*m1_r*m1_r)/np.power(m2_r, 1.5)
            skew[idx] = (m3_r - 3 * m1_r * m2 - m1_r * m1_r * m1_r) / np.power(m2, 1.5)
        return skew

    for key, params in params_dict.items():
        qs.append(pd.Series(skewness(params=params), index=kappa2s, name=key))

    ss_pdf = pd.concat(qs, axis=1)

    sns.lineplot(data=ss_pdf, dashes=False, ax=ax)
    ax.set_xlabel(f'$\kappa_{2}$')
    if title is not None:
        ax.set_title(title, fontsize=12, color='darkblue')


def plot_ss_kurtosis(params_dict=SS_PARAMS,
                     title: str = f'Excess kurtosis of log-returns as function of $\kappa_{2}$',
                     ax: plt.Subplot = None
                     ) -> None:

    kappa2s = np.linspace(0.5, 10.0, 100)
    qs = []

    def kurtosys(params: LogSvParams) -> np.ndarray:
        kurt = np.zeros_like(kappa2s)
        for idx, kappa2 in enumerate(kappa2s):
            nu = 2.0 * (kappa2 * params.theta-params.kappa1) / params.vartheta2 - 1.0
            q = 2.0 * params.kappa1 * params.theta / params.vartheta2
            b = 2.0 * kappa2 / params.vartheta2
            arg = 2.0*np.sqrt(q*b)
            kurt[idx] = 3.0*sps.kv(nu+4.0, arg)*sps.kv(nu, arg)/np.square(sps.kv(nu+2, arg)) - 3.0
        return kurt

    for key, params in params_dict.items():
        qs.append(pd.Series(kurtosys(params=params), index=kappa2s, name=key))

    ss_pdf = pd.concat(qs, axis=1)

    sns.lineplot(data=ss_pdf, dashes=False, ax=ax)
    ax.set_xlabel(f'$\kappa_{2}$')
    if title is not None:
        ax.set_title(title, fontsize=12, color='darkblue')


class UnitTests(Enum):
    PLOT_VOL_STEADY_STATE = 1
    PLOT_SS_PDF = 2
    PLOT_X_PDF = 3
    PLOT_KURT = 4
    JOINT_FIGURE = 5
    SKEWENESS = 6


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.PLOT_VOL_STEADY_STATE:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)
            plot_steady_state(title='Steady state distribution of volatility with $\kappa_{1}=4$',
                              ax=ax)

    elif unit_test == UnitTests.PLOT_SS_PDF:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(12, 6), tight_layout=True)
            plot_steady_state(ax=ax)

    elif unit_test == UnitTests.PLOT_X_PDF:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(12, 6), tight_layout=True)
            plot_steady_state_x(ax=ax)

    elif unit_test == UnitTests.PLOT_KURT:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(6, 6), tight_layout=True)
            plot_ss_kurtosis(ax=ax)

    elif unit_test == UnitTests.SKEWENESS:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(6, 6), tight_layout=True)
            plot_vol_skew(ax=ax)

    elif unit_test == UnitTests.JOINT_FIGURE:
        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 3, figsize=(18, 6), tight_layout=True)
            plot_steady_state(title='(A) Steady state distribution of the volatility',
                              ax=axs[0])
            plot_vol_skew(title=f'(B) Skeweness of volatility as function of $\kappa_{2}$',
                          ax=axs[1])
            plot_ss_kurtosis(title=f'(C) Excess kurtosis of log-returns as function of $\kappa_{2}$',
                             ax=axs[2])

            is_save = True
            if is_save:
                plot.save_fig(fig=fig, local_path='../../../docs/figures//', file_name='vol_steady_state')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_VOL_STEADY_STATE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum
import qis

from stochvolmodels.my_papers.volatility_models.load_data import fetch_ohlc_vol


def estimate_vol_beta(vol: pd.Series,
                      returns: pd.Series,
                      span: int = 33
                      ) -> pd.Series:
    # dvol = np.log(vol).diff(1).rename('dvol')
    dvol = vol.diff(1).rename('dvol')
    joint = pd.concat([dvol, returns], axis=1).dropna()
    vol_beta = qis.compute_one_factor_ewm_betas(x=joint[returns.name], y=joint['dvol'].to_frame(),
                                                span=span
                                                ).iloc[:, 0]
    return vol_beta


def plot_vol_beta(vol: pd.Series, returns: pd.Series, span: int = 33):
    vol_beta = estimate_vol_beta(vol=vol, returns=returns, span=span)

    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)
        qis.plot_time_series(df=vol_beta,
                             ax=ax)


class UnitTests(Enum):
    VOL_BETA = 1
    PLOT_VOL_BETA = 2


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.VOL_BETA:
        vol, returns = fetch_ohlc_vol(ticker='VIX')
        vol_beta = estimate_vol_beta(vol=vol, returns=returns)
        print(vol_beta)

    elif unit_test == UnitTests.PLOT_VOL_BETA:
        vol, returns = fetch_ohlc_vol(ticker='OVX')
        plot_vol_beta(vol=vol, returns=returns)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_VOL_BETA

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
figures 8 and 9
plot ODE system solutions in time
"""

import time
import numpy as np
import pandas as pd
import seaborn as sns
from scipy import linalg as sla
from numpy import linalg as la
from matplotlib import pyplot as plt
from enum import Enum
from scipy.integrate._ivp.ivp import OdeResult
from typing import Tuple, List

import stochvolmodels.utils.plots as plot
from stochvolmodels.pricers.logsv.affine_expansion import solve_ode_for_a, ExpansionOrder, func_a_ode_quadratic_terms, get_expansion_n
from stochvolmodels.pricers.logsv_pricer import LogSvParams


def plot_ode_sol_in_t(params: LogSvParams,
                      ttm: float,
                      ode_sol: OdeResult,
                      expansion_order: ExpansionOrder = ExpansionOrder.FIRST,
                      title: str = None,
                      headers: List[str] = None,
                      axs: List[plt.Subplot] = None
                      ) -> None:
    """
    plot real and imag parts of given solution
    """
    t = np.linspace(0.0, ttm, 300)
    z = ode_sol.sol(t)

    if expansion_order == ExpansionOrder.FIRST:
        y = params.sigma0 - params.theta
        ys = np.array([1.0, y, y * y])
        mgf = np.exp(z.T @ ys)
        data = {r'$\Re[A(\tau)]$': pd.DataFrame(data=np.real(z).T, index=t, columns=[r'$A^{(0)}$', r'$A^{(1)}$', r'$A^{(2)}$']),
                r'$\Im[A(\tau)]$': pd.DataFrame(data=np.imag(z).T, index=t, columns=[r'$A^{(0)}$', r'$A^{(1)}$', r'$A^{(2)}$']),
                r'$E^{[1]}(\tau)$': pd.concat([pd.Series(np.real(mgf), index=t, name=r'$\Re[E^{[1]}]$'),
                                         pd.Series(np.imag(mgf), index=t, name=r'$\Im[E^{[1]}]$')],
                                        axis=1)}

    elif expansion_order == ExpansionOrder.SECOND:
        y = params.sigma0 - params.theta
        ys = np.array([1.0, y, y * y, y ** 3, y ** 4])
        mgf = np.exp(z.T @ ys)
        data = {r'$\Re[A(\tau)]$': pd.DataFrame(data=np.real(z).T, index=t, columns=[r'$A^{(0)}$', r'$A^{(1)}$', r'$A^{(2)}$', r'$A^{(3)}$', r'$A^{(4)}$']),
                r'$\Im[A(\tau)]$': pd.DataFrame(data=np.imag(z).T, index=t, columns=[r'$A^{(0)}$', r'$A^{(1)}$', r'$A^{(2)}$', r'$A^{(3)}$', r'$A^{(4)}$']),
                r'$E^{[2]}(\tau)$': pd.concat([pd.Series(np.real(mgf), index=t, name=r'$\Re[E^{[2]}]$'),
                                         pd.Series(np.imag(mgf), index=t, name=r'$\Im[E^{[2]}]$')],
                                        axis=1)}

    else:
        raise TypeError(f"not implemented")

    if axs is None:
        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(3, 1, figsize=(8, 8), tight_layout=True)
            fig.suptitle(title, color='darkblue')
    for idx, (key, data) in enumerate(data.items()):
        sns.lineplot(data=data, ax=axs[idx])
        if headers is not None: # and idx == 0:
            title_ = f"{headers[idx]} {key}, {title}"
        else:
            title_ = key
        axs[idx].set_title(title_, color='darkblue')
        axs[idx].set(xlabel=r"$\tau$")


def plot_ode_solutions(params: LogSvParams,
                       ttm: float = 1,
                       is_stiff_solver: bool = True,
                       expansion_order: ExpansionOrder = ExpansionOrder.FIRST,
                       is_spot_measure: bool = True
                       ) -> plt.Figure:
    """
    solve and plot real and imag parts of solutions in ttm as function of phi
    """
    if is_spot_measure:
        real_part = -0.5
    else:
        real_part = 0.5

    is_1d = True
    if is_1d:
        phis = np.array([real_part + 2.0j], dtype=np.complex128)
        headers = [['(A)', '(B)', '(C)']]
    else:
        phis = np.array([real_part + 1.0j, real_part + 10.0j], dtype=np.complex128)
        headers = [['(A1)', '(A2)', '(A3)'], ['(B1)', '(B2)', '(B3)']]

    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(len(phis), 3, figsize=(18, 6), tight_layout=True)

    for idx, phi in enumerate(phis):
        tic = time.perf_counter()
        ode_sol = solve_ode_for_a(ttm=ttm,
                                  theta=params.theta,
                                  kappa1=params.kappa1,
                                  kappa2=params.kappa2,
                                  beta=params.beta,
                                  volvol=params.volvol,
                                  phi=phi,
                                  psi=0.0 + 0.0j,
                                  dense_output=True,
                                  expansion_order=expansion_order,
                                  is_stiff_solver=is_stiff_solver,
                                  is_spot_measure=is_spot_measure)
        toc = time.perf_counter()
        print(f"{toc - tic} secs to run solver")
        title = f"$\Phi$={np.real(phi):0.2f}+{np.imag(phi):0.2f}i"
        plot_ode_sol_in_t(params=params, ttm=ttm, ode_sol=ode_sol, expansion_order=expansion_order, title=title,
                          axs=axs[idx, :] if len(phis) > 1 else axs,
                          headers=headers[idx])
    plot.set_subplot_border(fig=fig, n_ax_rows=len(phis), n_ax_col=3)
    return fig


def solve_approximate_solutions(ttm: float,
                                theta: float,
                                kappa1: float,
                                kappa2: float,
                                beta: float,
                                volvol: float,
                                phi: np.complex128,
                                psi: np.complex128,
                                expansion_order: ExpansionOrder = ExpansionOrder.FIRST
                                ) -> np.ndarray:

    M, L, H = func_a_ode_quadratic_terms(theta=theta,
                                         kappa1=kappa1,
                                         kappa2=kappa2,
                                         beta=beta,
                                         volvol=volvol,
                                         phi=phi,
                                         psi=psi,
                                         expansion_order=expansion_order)
    L = np.transpose(L)
    term = (sla.expm(ttm*L) - la.pinv(L))*H
    return term


def plot_approximate_solutions(params: LogSvParams,
                               phi: np.complex128,
                               ttm: float = 1,
                               is_stiff_solver: bool = True,
                               expansion_order: ExpansionOrder = ExpansionOrder.FIRST
                               ) -> None:
    """
    solve and plot real and imag parts of solutions in ttm as function of phi
    """

    ttms = np.linspace(0.0, ttm, 100)
    apr_sol = np.ndarray((len(ttms), get_expansion_n(expansion_order)))
    for t, ttm in enumerate(ttms):
        apr_sol[t] = solve_approximate_solutions(ttm=ttm,
                                                 theta=params.theta,
                                                 kappa1=params.kappa1,
                                                 kappa2=params.kappa2,
                                                 beta=params.beta,
                                                 volvol=params.volvol,
                                                 phi=phi,
                                                 psi=0.0 + 0.0j,
                                                 expansion_order=expansion_order)[0]

    ode_sol = solve_ode_for_a(ttm=ttm,
                              theta=params.theta,
                              kappa1=params.kappa1,
                              kappa2=params.kappa2,
                              beta=params.beta,
                              volvol=params.volvol,
                              phi=phi,
                              psi=0.0 + 0.0j,
                              dense_output=True,
                              expansion_order=expansion_order,
                              is_stiff_solver=is_stiff_solver)
    z = ode_sol.sol(ttms)

    if expansion_order == ExpansionOrder.FIRST:
        columns = ['A0', 'A1', 'A2']
    elif expansion_order == ExpansionOrder.SECOND:
        columns = ['A0', 'A1', 'A2', 'A3', 'A4']
    else:
        raise TypeError(f"not implemented")

    apr_real = pd.DataFrame(data=np.real(apr_sol), index=ttms, columns=columns)
    apr_imag = pd.DataFrame(data=np.imag(apr_sol), index=ttms, columns=columns)
    sol_real = pd.DataFrame(data=np.real(z).T, index=ttms, columns=columns)
    sol_imag = pd.DataFrame(data=np.imag(z).T, index=ttms, columns=columns)

    real_data = {}
    imag_data = {}
    for column in columns:
        real_data[f"{column}: real"] = pd.concat([sol_real[column].rename('Solution'), apr_real[column].rename('Appr')], axis=1)
        imag_data[f"{column}: imag"] = pd.concat([sol_imag[column].rename('Solution'), apr_imag[column].rename('Appr')], axis=1)

    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(get_expansion_n(expansion_order), 2, figsize=(18, 6), tight_layout=True)
        title = f"phi={np.real(phi):0.2f}+{np.imag(phi):0.2f}i"
        fig.suptitle(title, color='darkblue')
        for idx, ((key1, data1), (key2, data2)) in enumerate(zip(real_data.items(), imag_data.items())):
            sns.lineplot(data=data1, ax=axs[idx, 0])
            axs[idx, 0].set_title(key1)
            sns.lineplot(data=data2, ax=axs[idx, 1])
            axs[idx, 1].set_title(key2)


def detect_ode_explosion(params: LogSvParams,
                         ttm: float = 1,
                         is_stiff_solver: bool = True,
                         expansion_order: ExpansionOrder = ExpansionOrder.FIRST
                         ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Detect an explosion is solution of ODE and check against having eigen value with positive real part in matrix
    of linear terms. Solution with value >=1e16 is considered an explosion
    """
    phis = np.array([-0.5 + 0.0j, -0.5 + 1.0j, -0.5 + 10.0j, -0.5 + 100.0j], dtype=np.complex128)
    t = np.linspace(0.0, ttm, 300)
    is_explosion = np.full(phis.size, False)
    pos_eigenvals = np.full(phis.size, False)

    for j_phi, phi in enumerate(phis):
        tic = time.perf_counter()
        ode_sol = solve_ode_for_a(ttm=ttm,
                                  theta=params.theta,
                                  kappa1=params.kappa1,
                                  kappa2=params.kappa2,
                                  beta=params.beta,
                                  volvol=params.volvol,
                                  phi=phi,
                                  psi=0.0 + 0.0j,
                                  dense_output=True,
                                  expansion_order=expansion_order,
                                  is_stiff_solver=is_stiff_solver)
        toc = time.perf_counter()
        print(f"{toc - tic} secs to run solver")
        # Calculate eigenvalues of linear term
        _, L, _ = func_a_ode_quadratic_terms(theta=params.theta,
                                             kappa1=params.kappa1,
                                             kappa2=params.kappa2,
                                             beta=params.beta,
                                             volvol=params.volvol,
                                             phi=phi,
                                             psi=0.0 + 0.0j,
                                             expansion_order=expansion_order,
                                             is_spot_measure=True)
        eigenvals = la.eigvals(L)
        if np.max(np.real(eigenvals)) > 0.0:
            pos_eigenvals[j_phi] = True
        z = ode_sol.sol(t)
        for zk in z:
            if np.max(zk) >= 1.0e16:
                is_explosion[j_phi] = True
                break
    return is_explosion, pos_eigenvals


class UnitTests(Enum):
    FIRST_ORDER = 1
    SECOND_ORDER = 2
    APPROXIMATION = 3


def run_unit_test(unit_test: UnitTests):

    # params = LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=4.0, beta=1.2, volvol=3.0)
    # params = LogSvParams(sigma0=0.5, theta=1.0, kappa1=4.0, kappa2=4.0, beta=1.0, volvol=1.0)
    params = LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8606, kappa2=4.7938, beta=0.1985, volvol=2.3690)

    params.assert_vol_moments_stability()
    params.print_vol_moments_stability()

    ttm = 1.0
    is_spot_measure = True

    is_save = False
    if unit_test == UnitTests.FIRST_ORDER:
        fig = plot_ode_solutions(params=params, ttm=ttm, expansion_order=ExpansionOrder.FIRST,
                                 is_spot_measure=is_spot_measure)
        if is_save:
            plot.save_fig(fig=fig, local_path='../../../docs/figures//', file_name="first_order_fig")

    elif unit_test == UnitTests.SECOND_ORDER:
        fig = plot_ode_solutions(params=params, ttm=ttm, expansion_order=ExpansionOrder.SECOND,
                                 is_spot_measure=is_spot_measure)
        if is_save:
            plot.save_fig(fig=fig, local_path='../../../docs/figures//', file_name="second_order_fig")

    elif unit_test == UnitTests.APPROXIMATION:
        plot_approximate_solutions(phi=-0.5 + 1j,
                                   params=params, ttm=ttm, expansion_order=ExpansionOrder.SECOND)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.SECOND_ORDER

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from scipy.special import betainc, gamma
from scipy.optimize import fsolve
from typing import Union
from enum import Enum

from stochvolmodels import TdistPricer, OptionChain


class UnitTests(Enum):
    SPY_FIT = 1
    GOLD_FIT = 2
    BTC_FIT = 3


def run_unit_test(unit_test: UnitTests):

    import seaborn as sns
    import qis as qis
    import stochvolmodels.data.test_option_chain as chains

    local_path = 'C://Users//artur//OneDrive//My Papers//Working Papers//Old Papers//Conditional Volatility Models. Zurich. May 2021//Figures//'

    if unit_test == UnitTests.SPY_FIT:
        option_chain = chains.get_spy_test_chain_data()
    elif unit_test == UnitTests.GOLD_FIT:
        option_chain = chains.get_gld_test_chain_data()
    elif unit_test == UnitTests.BTC_FIT:
        option_chain = chains.get_btc_test_chain_data()

    else:
        raise NotImplementedError

    tdist_pricer = TdistPricer()
    fit_params = tdist_pricer.calibrate_model_params_to_chain(option_chain=option_chain)

    with sns.axes_style('darkgrid'):
        fig, axs = plt.subplots(2, 2, figsize=(14, 12), tight_layout=True)
        axs = qis.to_flat_list(axs)

    for idx, (key, params) in enumerate(fit_params.items()):
        print(f"{key}: {params}")
        title = f"maturity-{key}: nu={params.nu:0.2f}, vol={params.vol:0.2f}, drift={params.drift:0.2%}"
        option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[key])
        tdist_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain0, params=params,
                                                 title=title, axs=[axs[idx]])

    qis.save_fig(fig, file_name=f"{unit_test.name.lower()}", local_path=local_path)
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.GOLD_FIT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import qis
from typing import Tuple
from scipy.optimize import minimize
from enum import Enum

import stochvolmodels.my_papers.volatility_models.ss_distribution_fit as ssd
from stochvolmodels.my_papers.volatility_models.load_data import fetch_ohlc_vol
from stochvolmodels.pricers.logsv_pricer import LogSvParams, LogSVPricer
from stochvolmodels.utils.funcs import set_seed


def compute_autocorr_power(alpha: float = 0.1, c: float = 1.0, num_lags: int = 20) -> np.ndarray:
    dts = np.arange(0, num_lags) / 260.0
    pf = 1.0 - c*np.power(dts, 2.0*alpha+1)
    return pf


def fit_autocorr_power(vol: pd.Series, num_lags: int = 60) -> Tuple[float, float]:

    # estimate empirical acf
    empirical = qis.compute_path_autocorr(a=vol.to_numpy(), num_lags=num_lags)

    def objective(pars: np.ndarray, args: np.ndarray) -> float:
        alpha, c = pars[0], pars[1]
        model_acfs = compute_autocorr_power(alpha=alpha, c=c, num_lags=num_lags)
        sse = np.nansum(np.square(model_acfs - empirical))
        return sse

    options = {'disp': True, 'ftol': 1e-8}
    p0 = np.array([0.1, 0.99])
    bounds = ((-0.5, 0.5), (0.01, 1.5))
    res = minimize(objective, p0, args=None, method='SLSQP', bounds=bounds, options=options)
    fit_alpha, fit_c = res.x[0], res.x[1]
    return fit_alpha, fit_c


def simulate_autocorr(params: LogSvParams,
                      brownians: np.ndarray = None,
                      nb_path: int = 1000,
                      nb_steps: int = 260,
                      num_lags: int = 20,
                      ttm: float = 10.0
                      ) -> np.ndarray:
    """
    use mc paths to compute paths of autocorrelation
    """
    logsv_pricer = LogSVPricer()
    if brownians is None:
        brownians = np.sqrt(1.0 / 260) * np.random.normal(0, 1, size=(nb_steps, nb_path))
    sigma_t, grid_t = logsv_pricer.simulate_vol_paths(params=params,
                                                      nb_path=nb_path,
                                                      nb_steps=nb_steps,
                                                      ttm=ttm,
                                                      brownians=brownians)
    acfs = qis.compute_path_autocorr(a=sigma_t, num_lags=num_lags)
    return acfs


def get_brownians(nb_steps: int, nb_path: int) -> np.ndarray:
    try:
        brownians = get_brownians.brownians
    except AttributeError:  # read static data and save for next call
        dt = 1.0 / 260.0
        brownians = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))
        get_brownians.brownians = brownians
    return brownians


def fit_autocorr_logsv(vol: pd.Series,
                       nb_path: int = 1000,
                       num_lags: int = 60,
                       ttm: float = 10.0
                       ) -> LogSvParams:
    """
    fit autocorrelation of log sv model using MC simulations
    """
    # fix brownians
    nb_steps = int(260*ttm)
    brownians = get_brownians(nb_steps=nb_steps, nb_path=nb_path)

    # estimate empirical acf
    empirical = qis.compute_path_autocorr(a=vol.to_numpy(), num_lags=num_lags)

    def unpack_pars(pars: np.ndarray) -> LogSvParams:
        kappa1, kappa2 = pars[0], pars[1]
        params = ssd.fit_distribution_log_sv_fixed_kappa(vol=vol, kappa1=kappa1, kappa2=kappa2)
        return params

    def objective(pars: np.ndarray, args: np.ndarray) -> float:
        params = unpack_pars(pars=pars)
        model_acfs = simulate_autocorr(params=params,
                                       brownians=brownians,
                                       nb_path=nb_path,
                                       nb_steps=nb_steps,
                                       num_lags=num_lags,
                                       ttm=ttm)
        model_acfs = np.mean(model_acfs, axis=1)
        sse = np.nansum(np.square(model_acfs - empirical))
        return sse

    options = {'disp': True, 'ftol': 1e-8}
    p0 = np.array([2.0, 2.0])
    bounds = ((0.2, 10), (0.2, 10))
    res = minimize(objective, p0, args=None, method='SLSQP', bounds=bounds, options=options)

    fit_params = unpack_pars(pars=res.x)
    return fit_params


def autocorr_fit_report_logsv(vol: pd.Series,
                              params: LogSvParams,
                              nb_path: int = 5000,
                              num_lags: int = 90,
                              ttm: float = 10.0,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:

    # estimate empirical acf
    index = range(0, num_lags)
    empirical = qis.compute_path_autocorr(a=vol.to_numpy(), num_lags=num_lags)
    empirical = pd.Series(empirical, index=index, name='Empirical')

    # simulated
    nb_steps = int(260*ttm)
    brownians = get_brownians(nb_steps=nb_steps, nb_path=nb_path)
    model_acfs = simulate_autocorr(params=params,
                                   brownians=brownians,
                                   nb_path=nb_path,
                                   nb_steps=nb_steps,
                                   num_lags=num_lags,
                                   ttm=ttm)
    model_acf = np.mean(model_acfs, axis=1)
    model_acf = pd.Series(model_acf, index=index, name='Log SV')

    # power
    alpha, c = fit_autocorr_power(vol=vol, num_lags=num_lags)
    pf = compute_autocorr_power(alpha=alpha, c=c, num_lags=num_lags)
    pf_power = pd.Series(pf, index=index, name='Rough ' + r'$\alpha$' + f"={alpha:0.2f}")

    # plot acfc
    acfs_df = pd.concat([empirical, model_acf, pf_power], axis=1)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)

    qis.plot_line(acfs_df,
                  linestyles=['dotted', 'solid', 'dashed'],
                  ax=ax,
                  legend_loc='upper center',
                  xlabel='Lag',
                  **kwargs)


class UnitTests(Enum):
    EMPIRICAL_AUTOCORR = 1
    AUTOCORR_POWER = 2
    FIT_AUTOCORR_LOGSV = 3
    FIT_AUTOCORR_HESTON = 4
    FIT_REPORT = 5


def run_unit_test(unit_test: UnitTests):

    set_seed(3)
    np.random.seed(3)

    vix_log_params = LogSvParams(sigma0=0.19928505844247962, theta=0.19928505844247962, kappa1=1.2878835150774184,
                                 kappa2=1.9267876555824357, beta=0.0, volvol=0.7210463316739526)

    move_log_params = LogSvParams(sigma0=0.9109917133860931, theta=0.9109917133860931,
                                  kappa1=0.1, kappa2=0.41131244621275886, beta=0.0, volvol=0.3564212939473691)

    ovx_log_params = LogSvParams(sigma0=0.3852514800317871, theta=0.3852514800317871, kappa1=2.7774564907918027,
                                 kappa2=2.2351296851221107, beta=0.0, volvol=0.8344408577025486)

    btc_log_params = LogSvParams(sigma0=0.7118361434192538, theta=0.7118361434192538,
                                 kappa1=2.214702576955766, kappa2=2.18028273418397, beta=0.0, volvol=0.921487415907961)

    eth_log_params = LogSvParams(sigma0=0.8657438901704476, theta=0.8657438901704476, kappa1=1.955809653686808,
                                 kappa2=1.978367101612294, beta=0.0, volvol=0.8484117641903834)

    nb_path: int = 10000
    num_lags: int = 120
    ttm: float = 10.0

    if unit_test == UnitTests.EMPIRICAL_AUTOCORR:
        vol, returns = fetch_ohlc_vol(ticker='VIX')
        empirical = qis.compute_path_autocorr(a=vol.to_numpy(), num_lags=num_lags)
        print(empirical)
        qis.plot_line(df=pd.Series(empirical))

        dvols = vol.diff(1)
        vol_returns = qis.compute_path_corr(a1=dvols.to_numpy(), a2=returns.to_numpy(), num_lags=num_lags)[1:]
        print(vol_returns)
        qis.plot_line(df=pd.Series(vol_returns))

    elif unit_test == UnitTests.AUTOCORR_POWER:
        vol, returns = fetch_ohlc_vol(ticker='OVX')
        alpha, c = fit_autocorr_power(vol=vol)
        print(f"alpha={alpha}, c={c}")
        pf = compute_autocorr_power(alpha=alpha, num_lags=60)
        print(pf)

    elif unit_test == UnitTests.FIT_AUTOCORR_LOGSV:
        vol, returns = fetch_ohlc_vol(ticker='ETH')
        fit_params = fit_autocorr_logsv(vol=vol, nb_path=nb_path, num_lags=num_lags, ttm=ttm)
        print(f"fit_params={fit_params}")
        qis.plot_time_series(df=vol)
        autocorr_fit_report_logsv(params=fit_params, vol=vol, nb_path=nb_path, num_lags=num_lags, ttm=ttm)
        ssd.plot_estimated_svs(vol=vol, logsv_params=fit_params, heston_params=None, bins=50)

    elif unit_test == UnitTests.FIT_REPORT:
        vol, returns = fetch_ohlc_vol(ticker='VIX')
        autocorr_fit_report_logsv(params=vix_log_params,
                                  vol=vol,
                                  nb_path=nb_path,
                                  num_lags=120,
                                  ttm=ttm)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.FIT_AUTOCORR_LOGSV

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)


"""
figures 2 and 3 in the logsv_model_wtih_quadratic_drift
plot admissible regions for model parameters
"""

import numpy as np
import matplotlib.pyplot as plt
from enum import Enum

import stochvolmodels.utils.plots as plot


def lognormal_combined(vartheta_min=0.5,
                       vartheta_max=3.0,
                       beta_min=-2.5,
                       beta_max=2.5,
                       kappa2s=[3.0, 0.0]):
    hatch1 = '\\\\\\\\'
    hatch2 = '////'
    vartheta = np.linspace(vartheta_min, vartheta_max, 100)
    nb_kappas = len(kappa2s)
    moment = 2
    num = 97

    fig, ax = plt.subplots(1, nb_kappas, figsize=(4 * nb_kappas, 3), tight_layout=True)
    for idx, kappa2 in enumerate(kappa2s):
        # plot for bounds ensuring martingale property under spot and inverse spot
        beta_spot_meas = np.maximum(kappa2, beta_min)
        beta_spot_meas = np.ones_like(vartheta) * beta_spot_meas
        beta_inv_meas = np.maximum(0.5 * kappa2, beta_min)
        beta_inv_meas = np.ones_like(vartheta) * beta_inv_meas
        outline, = ax[idx].plot(vartheta, beta_spot_meas, color='black', linewidth=0.8)
        ax[idx].fill_between(vartheta, beta_min, beta_spot_meas,
                                edgecolor='black', hatch=hatch1, label='MMA', facecolor='none')
        outline, = ax[idx].plot(vartheta, beta_inv_meas, color='black', linewidth=0.8)
        ax[idx].fill_between(vartheta, beta_min, beta_inv_meas,
                                edgecolor='grey', hatch=hatch2, label='Inverse', facecolor='none')
        ax[idx].set_ylim(beta_min, beta_max, auto=True)
        ax[idx].set_title(f"({chr(num).upper()}): $\kappa_2={kappa2}$")
        num = num + 1
        # # Right plot for bounds ensuring existence of second moment under spot and inverse spot
        # m = moment
        # beta_spot_meas = np.maximum((kappa2 - vartheta * np.sqrt(m ** 2 - m)) / m, beta_min)
        # beta_inv_meas = np.maximum((kappa2 - vartheta * np.sqrt(m ** 2 - m)) / (m + 1.0), beta_min)
        # outline, = ax[idx, 1].plot(vartheta, beta_spot_meas, color='black', linewidth=0.8)
        # ax[idx, 1].fill_between(vartheta, beta_min, beta_spot_meas,
        #                         edgecolor='black', hatch=hatch1, label='Spot', facecolor='none')
        # outline, = ax[idx, 1].plot(vartheta, beta_inv_meas, color='black', linewidth=0.8)
        # ax[idx, 1].fill_between(vartheta, beta_min, beta_inv_meas,
        #                         edgecolor='grey', hatch=hatch2, label='Inverse', facecolor='none')
        # ax[idx, 1].set_ylim(beta_min, beta_max, auto=True)
        # ax[idx, 1].set_title(f"({chr(num).upper()}): $\kappa_2={kappa2}$")
        # num = num + 1
    for i in range(nb_kappas):
        # for j in range(2):
        ax[i].legend()
        ax[i].set(xlabel=r"$\vartheta$", ylabel=r"$\beta$")

    plot.save_fig(fig=fig, local_path='../../../docs/figures//',
                  file_name='logsv_regions')


def heston_exp_ou_combined(vartheta_min=0.5,
                           vartheta_max=3.0,
                           rho_min=-1.0,
                           rho_max=1.0,
                           kappa=1,
                           theta=1):
    hatch1 = '\\\\\\\\'
    hatch2 = '////'
    vartheta = np.linspace(vartheta_min, vartheta_max, 100)

    fig, ax = plt.subplots(1, 2, figsize=(10, 3), tight_layout=True)
    # Left plot for bounds ensuring existence of second moment under spot and inverse spot
    kappa_adj_inv = np.maximum(kappa / vartheta, rho_min)
    outline, = ax[0].plot(vartheta, kappa_adj_inv, color='black', linewidth=0.8)
    ax[0].fill_between(vartheta, rho_min, kappa_adj_inv,
                       edgecolor='black', hatch=hatch1, label=r"$\kappa>\rho\vartheta$",
                       facecolor='none')
    critical_val = np.sqrt(2.0 * kappa * theta)
    ax[0].axvspan(vartheta_min, critical_val, edgecolor='black', hatch=hatch2, label='Feller',
                  facecolor='none')
    ax[0].set_ylim(rho_min, rho_max, auto=True)
    ax[0].legend()
    ax[0].set(xlabel=r"$\vartheta$", ylabel=r"$\rho$")
    ax[0].set_title(f"(A) Heston model")
    # Right plot for bounds ensuring existence of second moment under spot and inverse spot
    kappa2 = 0
    rho_spot_meas = np.maximum(kappa2 / vartheta, rho_min)
    rho_inv_meas = np.maximum(0.5 * kappa2 / vartheta, rho_min)
    outline, = ax[1].plot(vartheta, rho_spot_meas, color='black', linewidth=0.8)
    ax[1].fill_between(vartheta, rho_min, rho_spot_meas,
                       edgecolor='black', hatch=hatch1, label='MMA',
                       facecolor='none')
    outline, = ax[1].plot(vartheta, rho_inv_meas, color='black', linewidth=0.8)
    ax[1].fill_between(vartheta, rho_min, rho_inv_meas,
                       edgecolor='grey', hatch=hatch2, label='Inverse',
                       facecolor='none')
    ax[1].set_ylim(rho_min, rho_max, auto=True)
    ax[1].legend()
    ax[1].set(xlabel=r"$\vartheta$", ylabel=r"$\rho$")
    ax[1].set_title(f"(B) Exp-OU model")

    plot.save_fig(fig=fig, local_path='../../../docs/figures//',
                  file_name='heston_exp_ou_combined')


class UnitTests(Enum):
    LOGNORMAL_SV_COMBINED = 1
    HESTON_EXP_OU_COMBINED = 2


def run_unit_test(unit_test: UnitTests):
    if unit_test == UnitTests.LOGNORMAL_SV_COMBINED:
        lognormal_combined(vartheta_min=0.5,
                           vartheta_max=3.0,
                           kappa2s=[0.0, 1.0])

    elif unit_test == UnitTests.HESTON_EXP_OU_COMBINED:
        heston_exp_ou_combined(vartheta_min=0.5,
                               vartheta_max=3.0)

    plt.show()


if __name__ == '__main__':
    unit_test = UnitTests.LOGNORMAL_SV_COMBINED
    run_unit_test(unit_test=unit_test)




import numpy as np
from numba import njit
from stochvolmodels import npdf, infer_bsm_implied_vol


@njit
def compute_normal_pdf(x: np.ndarray):
    dx = x[1] - x[0]
    return dx*npdf(x)


@njit
def value_under_q_kernel(b: float, pdf: np.ndarray, x: np.ndarray, payoff: np.ndarray, forward: float = 1.0):
    c = -0.5 + (2.0*b+1.0)*np.log(forward)
    norm = np.exp(0.5*np.square(c)/(2.0*b+1))/np.sqrt(2.0*b+1)
    q_payoff = np.sum(pdf*np.exp(c*x-b*np.square(x))*payoff) / norm
    return q_payoff


@njit
def value_payoff(pdf: np.ndarray, payoff: np.ndarray):
    return np.sum(pdf*payoff)


x = np.linspace(-5.0, 5.0, 20000)
pdf = compute_normal_pdf(x)

print(f"sum={np.sum(pdf)}, mean={np.sum(x * pdf)}, std={np.sqrt(np.sum(np.square(x) * pdf) - np.square(np.sum(x * pdf)))}")

payoff = np.exp(x)
q_payoff = value_under_q_kernel(b=2.0, pdf=pdf, x=x, payoff=payoff, forward=1)
print(f"q_payoff={q_payoff}")

strikes = np.linspace(0.2, 2.0, 21)

values, model_vols = np.zeros_like(strikes), np.zeros_like(strikes)
values_q, model_vols_q = np.zeros_like(strikes), np.zeros_like(strikes)
for idx, strike in enumerate(strikes):
    spot = np.exp(x-0.5)
    payoff = np.maximum(spot-strike, 0.0)
    model_price = value_payoff(pdf=pdf, payoff=payoff)
    values[idx] = model_price
    model_vols[idx] = infer_bsm_implied_vol(forward=1.0, ttm=1.0, given_price=model_price, strike=strike, optiontype='C')
    payoff = np.maximum(np.exp(x)-strike, 0.0)
    model_price_q = value_under_q_kernel(b=0.25, pdf=pdf, x=x, payoff=payoff)
    values_q[idx] = model_price_q
    model_vols_q[idx] = infer_bsm_implied_vol(forward=1.0, ttm=1.0, given_price=model_price_q, strike=strike, optiontype='C')

print(f"values={values}")
print(f"values_q={values_q}")
print(f"model_vols={model_vols}")
print(f"model_vols_q={model_vols_q}")



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from enum import Enum


from stochvolmodels import (compute_bsm_vanilla_grid_deltas,
                            compute_bsm_forward_grid_prices,
                            compute_bsm_vanilla_price,
                            compute_bsm_vanilla_delta)


def compare_net_deltas(ttm: float,
                       forward: float,
                       vol: float,
                       strike_level: float = 1.0,
                       optiontype: str = 'C',
                       ax: plt.Subplot = None,
                       **kwargs
                       ):

    spot_drid = np.linspace(0.7*forward, 1.3*forward, 1000)
    strike = strike_level * forward

    option_prices = compute_bsm_forward_grid_prices(ttm=ttm, forwards=spot_drid, strike=strike, vol=vol, optiontype=optiontype)
    option_deltas = compute_bsm_vanilla_grid_deltas(ttm=ttm, forwards=spot_drid, strike=strike, vol=vol, optiontype=optiontype)
    option_net_delta = option_deltas - option_prices / spot_drid
    option_deltas = pd.Series(option_deltas, index=spot_drid, name='Black Delta')
    option_net_delta = pd.Series(option_net_delta, index=spot_drid, name='Net Delta')
    deltas = pd.concat([option_deltas, option_net_delta], axis=1)

    qis.plot_line(df=deltas,
                  xvar_format='{:,.0f}',
                  ylabel='BTC price',
                  ax=ax,
                  **kwargs)


def compare_pnl(ttm: float,
                forward: float,
                vol: float,
                strike_level: float = 1.0,
                optiontype: str = 'C',
                is_btc_pnl: bool = True,
                ax: plt.Subplot = None,
                **kwargs
                ):

    spot_drid = np.linspace(0.7*forward, 1.3*forward, 10000)
    returns_grid = spot_drid / forward - 1.0
    strike = strike_level*forward

    option_price0 = compute_bsm_vanilla_price(ttm=ttm, forward=forward, strike=strike, vol=vol, optiontype=optiontype)
    option_delta0 = compute_bsm_vanilla_delta(ttm=ttm, forward=forward, strike=strike, vol=vol, optiontype=optiontype)
    option_net_delta0 = option_delta0 - option_price0 / forward

    # price return
    inverse_price_return = (spot_drid-forward) / spot_drid
    dt = 1.0 / 365.0
    option_prices = compute_bsm_forward_grid_prices(ttm=ttm-dt, forwards=spot_drid, strike=strike, vol=vol, optiontype=optiontype)
    option_pnl_btc = (option_price0/forward - option_prices/spot_drid)

    # black p&l
    pnl_btc = option_pnl_btc + option_delta0 * inverse_price_return
    if not is_btc_pnl:
        pnl_btc = pnl_btc * spot_drid
    pnl_btc_positive = spot_drid[pnl_btc >= 0.0]
    lower_be = pnl_btc_positive[0] / forward - 1.0
    upper_be = pnl_btc_positive[-1] / forward - 1.0
    pnl_btc = pd.Series(pnl_btc, index=returns_grid, name=f"Black Delta: breakevens=({lower_be:0.2%}, {upper_be:0.2%})")

    # net p&l
    pnl_btc_net_delta = option_pnl_btc + option_net_delta0 * inverse_price_return
    if not is_btc_pnl:
        pnl_btc_net_delta = pnl_btc_net_delta * spot_drid
    pnl_btc_net_delta_positive = spot_drid[pnl_btc_net_delta >= 0.0]
    lower_be = pnl_btc_net_delta_positive[0] / forward - 1.0
    upper_be = pnl_btc_net_delta_positive[-1] / forward - 1.0
    pnl_btc_net_delta = pd.Series(pnl_btc_net_delta, index=returns_grid, name=f"Net Delta   : breakevens=({lower_be:0.2%}, {upper_be:0.2%})")

    pnls = pd.concat([pnl_btc, pnl_btc_net_delta], axis=1)

    if is_btc_pnl:
        ylabel = 'BTC P&L'
        yvar_format = '{:,.2f}'
    else:
        ylabel = 'USD P&L'
        yvar_format = '{:,.0f}'

    qis.plot_line(df=pnls,
                  xvar_format='{:,.1%}',
                  yvar_format=yvar_format,
                  ylabel=ylabel,
                  xlabel='BTC % change',
                  ax=ax,
                  **kwargs)


class UnitTests(Enum):
    DELTA_COMP = 1
    PNL_COMP = 2


def run_unit_test(unit_test: UnitTests):

    LOCAL_PATH = "C://Users//artur//OneDrive//My Papers//Working Papers//Crypto Options. Zurich. Oct 2022//FinalFigures//"

    kwargs = dict(fontsize=14, framealpha=0.9)

    ttm = 7.0 / 365.0

    if unit_test == UnitTests.DELTA_COMP:

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 2, figsize=(14, 7), tight_layout=True)
            compare_net_deltas(ttm=ttm, forward=50000, vol=0.6, optiontype='C',
                               strike_level=1.0,
                               title='(A) Call Delta for K=100%*S_0',
                               ax=axs[0],
                               **kwargs)
            compare_net_deltas(ttm=ttm, forward=50000, vol=0.6, optiontype='P',
                               strike_level=1.0,
                               title='(B) Put Delta for K=100%*S_0',
                               ax=axs[1],
                               **kwargs)

        is_save = True
        if is_save:
            qis.save_fig(fig, file_name='delta_comp', local_path=LOCAL_PATH)
            qis.save_fig(fig, file_name='delta_comp',  file_type=qis.FileTypes.EPS, dpi=1200, local_path=LOCAL_PATH)

    elif unit_test == UnitTests.PNL_COMP:

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 2, figsize=(14, 7), tight_layout=True)
            compare_pnl(ttm=ttm, forward=50000, vol=0.6, optiontype='C',
                        title='(A) ATM Call K=100%*F_0',
                        y_limits=(-0.3, 0.05),
                        ax=axs[0],
                        **kwargs)
            compare_pnl(ttm=ttm, forward=50000, vol=0.6, optiontype='C',
                        title='(B) ITM Call K=90%*F_0',
                        strike_level=0.9,
                        is_btc_pnl=True,
                        y_limits=(-0.3, 0.05),
                        ax=axs[1],
                        **kwargs)

        is_save = True
        if is_save:
            qis.save_fig(fig, file_name='pnl_comp', local_path=LOCAL_PATH)
            qis.save_fig(fig, file_name='pnl_comp',  file_type=qis.FileTypes.EPS, dpi=1200, local_path=LOCAL_PATH)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.DELTA_COMP

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List
from enum import Enum
import qis as qis

from stochvolmodels import (compute_vanilla_price_tdist,
                            infer_bsm_ivols_from_slice_prices,
                            infer_normal_ivols_from_slice_prices)

from stochvolmodels.pricers.analytic.tdist import (imply_drift_tdist,
                                                   compute_default_prob_tdist,
                                                   compute_forward_tdist)


def plot_implied_drift_forward_defaultp(spot: float = 1.0, vol: float = 0.2, nu: float = 3.0 ) -> plt.Figure:
    ttms = np.linspace(0.004, 1.0, 20)
    rf_rates = np.linspace(0.0, 0.05, 6)
    mus_ttm = {}
    forwards_ttm = {}
    default_prob_ttm = {}
    for rf_rate in rf_rates:
        mus = np.zeros_like(ttms)
        forwards = np.zeros_like(ttms)
        default_prob = np.zeros_like(ttms)
        for idx, ttm in enumerate(ttms):
            mus[idx] = imply_drift_tdist(rf_rate=rf_rate, vol=vol, nu=nu, ttm=ttm)
            forwards[idx] = compute_forward_tdist(spot=spot, rf_rate=rf_rate, vol=vol, nu=nu, ttm=ttm)
            default_prob[idx] = compute_default_prob_tdist(rf_rate=rf_rate, vol=vol, nu=nu, ttm=ttm)
        mus_ttm[f"rf_rate={rf_rate:,.2%}"] = pd.Series(mus, index=ttms)
        forwards_ttm[f"rf_rate={rf_rate:,.2%}"] = pd.Series(forwards, index=ttms)
        default_prob_ttm[f"rf_rate={rf_rate:,.2%}"] = pd.Series(default_prob, index=ttms)

    mus_ttm = pd.DataFrame.from_dict(mus_ttm, orient='columns')
    forwards_ttm = pd.DataFrame.from_dict(forwards_ttm, orient='columns')
    default_prob_ttm = pd.DataFrame.from_dict(default_prob_ttm, orient='columns')

    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(3, 1, figsize=(14, 14), tight_layout=True)
        qis.plot_line(df=mus_ttm,
                      yvar_format='{:,.2%}',
                      xvar_format='{:,.2f}',
                      xlabel='ttm',
                      title='(A) Implied Drift',
                      ax=axs[0])
        axs[0].set_xticklabels('')
        qis.plot_line(df=forwards_ttm,
                      yvar_format='{:,.2f}',
                      xvar_format='{:,.2f}',
                      xlabel='ttm',
                      title='(B) Model Forward',
                      ax=axs[1])
        axs[1].set_xticklabels('')
        qis.plot_line(df=default_prob_ttm,
                      yvar_format='{:,.2%}',
                      xvar_format='{:,.2f}',
                      xlabel='ttm',
                      title='(C) Model Default prob',
                      ax=axs[2])

    return fig


def plot_tdist_ivols_vs_bsm_normal(spot: float = 1.0,
                                   vol: float = 0.5,
                                   nu: float = 2.5,
                                   ttm: float = 1.0 / 12.0,
                                   rf_rate: float = 0.0,
                                   ax: plt.Subplot = None
                                   ) -> None:
    strikes = np.linspace(0.5, 1.5, 40)
    optiontypes_ttm = np.where(strikes <= 1.0, 'P', 'C')
    prices = compute_vanilla_price_tdist(spot=1.0, strikes=strikes, optiontypes=optiontypes_ttm, ttm=ttm, vol=vol, nu=nu, rf_rate=rf_rate)
    discfactor = np.exp(-rf_rate*ttm)
    forward = np.exp(rf_rate*ttm) * spot
    bsm_vols = infer_bsm_ivols_from_slice_prices(ttm=ttm, forward=forward, strikes=strikes,
                                                 optiontypes=optiontypes_ttm,
                                                 model_prices=prices,
                                                 discfactor=discfactor)
    bsm_vols = pd.Series(bsm_vols, index=strikes, name='BSM implied vol')
    normal_vols = infer_normal_ivols_from_slice_prices(ttm=ttm, forward=forward, strikes=strikes,
                                                       optiontypes=optiontypes_ttm,
                                                       model_prices=prices,
                                                       discfactor=discfactor)
    normal_vols = pd.Series(normal_vols, index=strikes, name='Normal implied vol')
    vols = pd.concat([bsm_vols, normal_vols], axis=1)
    qis.plot_line(df=vols, title=f"t-distribution implied vols, nu = {nu:0.2f}", ax=ax)


def plot_tdist_ivols_nu(spot: float = 1.0,
                        vol: float = 0.5,
                        ttm: float= 1.0 / 12.0,
                        nus: List[float] = [2.5, 3.0, 4.0, 5.0, 10.0, 20.0],
                        rf_rate: float = 0.00,
                        ax: plt.Subplot = None
                        ) -> None:

    forward = spot * np.exp(ttm*rf_rate)
    discfactor = np.exp(-ttm*rf_rate)
    strikes = np.linspace(0.5, 1.5, 100)
    optiontypes_ttm = np.where(strikes <= 1.0, 'P', 'C')
    bsm_vols = {}
    for nu in nus:
        prices = compute_vanilla_price_tdist(spot=spot, strikes=strikes, optiontypes=optiontypes_ttm, ttm=ttm, vol=vol, nu=nu,
                                             rf_rate=rf_rate)
        print(prices)
        bsm_vols[f"nu={nu:0.2f}"] = infer_bsm_ivols_from_slice_prices(ttm=ttm, forward=forward, discfactor=discfactor,
                                                                      strikes=strikes,
                                                                      optiontypes=optiontypes_ttm,
                                                                      model_prices=prices)
    bsm_vols = pd.DataFrame.from_dict(bsm_vols, orient='columns')
    bsm_vols.index = strikes
    qis.plot_line(df=bsm_vols,
                  title=f"t-distribution implied BSM vols, ttm={ttm:0.2f}",
                  xvar_format='{:,.0%}',
                  yvar_format='{:,.0%}',
                  xlabel='% strike',
                  ylabel='Implied vol',
                  ax=ax)


def plot_tdist_ivols_vol(vols: List[float] = [0.2, 0.3, 0.4, 0.8],
                         ttm: float= 1.0 / 12.0,
                         nu: float = 2.5,
                         ax: plt.Subplot = None
                         ) -> None:

    strikes = np.linspace(0.5, 1.5, 100)
    optiontypes_ttm = np.where(strikes <= 1.0, 'P', 'C')
    bsm_vols = {}
    for vol in vols:
        prices = compute_vanilla_price_tdist(spot=1.0, strikes=strikes, optiontypes=optiontypes_ttm, ttm=ttm, vol=vol, nu=nu)
        bsm_vols[f"vol={vol:0.2f}"] = infer_bsm_ivols_from_slice_prices(ttm=ttm, forward=1.0, discfactor=1.0, strikes=strikes,
                                                                      optiontypes=optiontypes_ttm,
                                                                      model_prices=prices)
    bsm_vols = pd.DataFrame.from_dict(bsm_vols, orient='columns')
    bsm_vols.index = strikes
    qis.plot_line(df=bsm_vols,
                  title=f"t-distribution mplied BSM vols, ttm={ttm:0.2f}",
                  xvar_format='{:,.0%}',
                  yvar_format='{:,.0%}',
                  xlabel='% strike',
                  ylabel='Implied vol',
                  ax=ax)


class UnitTests(Enum):
    PLOT_IMPLIED_DRIFT_FORWARD_DEFAULTPROB = 1
    PLOT_IMPLIED_VOLS_VS_BSM_NORMAL = 2
    PLOT_IVOLS_NU = 3
    PLOT_IVOLS_VOL = 4


def run_unit_test(unit_test: UnitTests):

    local_path = 'C://Users//artur//OneDrive//My Papers//Working Papers//Old Papers//Conditional Volatility Models. Zurich. May 2021//Figures//'

    if unit_test == UnitTests.PLOT_IMPLIED_DRIFT_FORWARD_DEFAULTPROB:
        fig = plot_implied_drift_forward_defaultp()
        qis.save_fig(fig=fig, file_name='mus', local_path=local_path)

    elif unit_test == UnitTests.PLOT_IMPLIED_VOLS_VS_BSM_NORMAL:
        plot_tdist_ivols_vs_bsm_normal(vol=0.5, nu=2.5)
        plot_tdist_ivols_vs_bsm_normal(vol=0.5, nu=5.0)

    elif unit_test == UnitTests.PLOT_IVOLS_NU:
        rf_rate = 0.0
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(1, 2, figsize=(14, 6), tight_layout=True)
            plot_tdist_ivols_nu(vol=0.2, ttm=5.0/252.0, rf_rate=rf_rate, ax=axs[0])
            plot_tdist_ivols_nu(vol=0.2, ttm=1.0/12.0, rf_rate=rf_rate, ax=axs[1])
            qis.align_y_limits_axs(axs)
        qis.save_fig(fig=fig, file_name='vols_in_nu', local_path=local_path)

    elif unit_test == UnitTests.PLOT_IVOLS_VOL:
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(1, 2, figsize=(14, 6), tight_layout=True)
            plot_tdist_ivols_vol(nu=2.5, ttm=5.0/252.0, ax=axs[0])
            plot_tdist_ivols_vol(nu=2.5, ttm=1.0/12.0, ax=axs[1])
            qis.align_y_limits_axs(axs)
        qis.save_fig(fig=fig, file_name='vols_in_vol', local_path=local_path)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_IVOLS_NU

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
run few unit test to illustrate implementation of log-normal sv model analytics
"""

import numpy as np
import matplotlib.pyplot as plt
from enum import Enum

import stochvolmodels as sv
from stochvolmodels import LogSVPricer, LogSvParams, OptionChain, LogsvModelCalibrationType


class UnitTests(Enum):
    COMPUTE_MODEL_PRICES = 1
    PLOT_MODEL_IMPLIED_VOLS = 2
    PLOT_MODEL_VOLS_IN_PARAMS = 3
    COMPARE_MODEL_VOLS_TO_MC = 4
    PLOT_FIT_TO_BITCOIN_OPTION_CHAIN = 5
    CALIBRATE_MODEL_TO_BTC_OPTIONS = 6


def run_unit_test(unit_test: UnitTests):

    # instance of pricer
    logsv_pricer = LogSVPricer()
    
    # define model params    
    params = LogSvParams(sigma0=1.0, theta=1.0, kappa1=5.0, kappa2=5.0, beta=0.2, volvol=2.0)

    if unit_test == UnitTests.COMPUTE_MODEL_PRICES:
        # 1. one price
        model_price, vol = logsv_pricer.price_vanilla(params=params,
                                                      ttm=0.25,
                                                      forward=1.0,
                                                      strike=1.0,
                                                      optiontype='C')
        print(f"price={model_price:0.4f}, implied vol={vol: 0.2%}")

        # 2. prices for slices
        model_prices, vols = logsv_pricer.price_slice(params=params,
                                                      ttm=0.25,
                                                      forward=1.0,
                                                      strikes=np.array([0.9, 1.0, 1.1]),
                                                      optiontypes=np.array(['P', 'C', 'C']))
        print([f"{p:0.4f}, implied vol={v: 0.2%}" for p, v in zip(model_prices, vols)])

        # 3. prices for option chain with uniform strikes
        option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                                     ids=np.array(['1m', '3m']),
                                                     strikes=np.linspace(0.9, 1.1, 3))
        model_prices, vols = logsv_pricer.compute_chain_prices_with_vols(option_chain=option_chain, params=params)
        print(model_prices)
        print(vols)

    elif unit_test == UnitTests.PLOT_MODEL_IMPLIED_VOLS:
        # define uniform option chain
        option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                                     ids=np.array(['1m', '3m']),
                                                     strikes=np.linspace(0.5, 1.5, 21))
        logsv_pricer.plot_model_ivols(option_chain=option_chain,
                                      params=params)

    elif unit_test == UnitTests.PLOT_MODEL_VOLS_IN_PARAMS:

        # define uniform option chain
        option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                                     ids=np.array(['1m', '3m']),
                                                     strikes=np.linspace(0.5, 1.5, 21))

        # define parameters for bootstrap
        params_dict = {'kappa2=5': LogSvParams(sigma0=1.0, theta=1.0, kappa1=5.0, kappa2=5.0, beta=0.2, volvol=2.0),
                       'kappa2=10': LogSvParams(sigma0=1.0, theta=1.0, kappa1=5.0, kappa2=10.0, beta=0.2, volvol=2.0)}

        # get slice for illustration
        option_slice = option_chain.get_slice(id='1m')
        logsv_pricer.plot_model_slices_in_params(option_slice=option_slice,
                                                 params_dict=params_dict)

    elif unit_test == UnitTests.COMPARE_MODEL_VOLS_TO_MC:
        btc_option_chain = sv.get_btc_test_chain_data()
        uniform_chain_data = OptionChain.to_uniform_strikes(obj=btc_option_chain, num_strikes=31)
        btc_calibrated_params = LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8609, kappa2=4.7940, beta=0.1988, volvol=2.3694)

        logsv_pricer.plot_model_ivols_vs_mc(option_chain=uniform_chain_data,
                                            params=btc_calibrated_params,
                                            nb_path=100000)

        logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=uniform_chain_data,
                                                           params=btc_calibrated_params,
                                                           nb_path=100000)

    elif unit_test == UnitTests.PLOT_FIT_TO_BITCOIN_OPTION_CHAIN:

        btc_option_chain = sv.get_btc_test_chain_data()
        btc_calibrated_params = LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8609, kappa2=4.7940, beta=0.1988, volvol=2.3694)
        logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=btc_option_chain,
                                                 params=btc_calibrated_params)

    elif unit_test == UnitTests.CALIBRATE_MODEL_TO_BTC_OPTIONS:
        btc_option_chain = sv.get_btc_test_chain_data()
        params0 = LogSvParams(sigma0=0.8, theta=1.0, kappa1=2.21, kappa2=2.18, beta=0.15, volvol=2.0)
        btc_calibrated_params = logsv_pricer.calibrate_model_params_to_chain(option_chain=btc_option_chain,
                                                                             params0=params0,
                                                                             model_calibration_type=LogsvModelCalibrationType.PARAMS4,

                                                                             constraints_type=sv.ConstraintsType.INVERSE_MARTINGALE)
        print(btc_calibrated_params)
        logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=btc_option_chain,
                                                 params=btc_calibrated_params)
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.COMPARE_MODEL_VOLS_TO_MC

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
run few unit test to illustrate implementation of log-normal sv model analytics
"""

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum

from stochvolmodels import LogSVPricer, LogSvParams, OptionChain


def plot_skews():
    logsv_pricer = LogSVPricer()
    option_chain = OptionChain.get_uniform_chain(ttms=np.array([14.0/365.0]),
                                                 ids=np.array(['2w']),
                                                 strikes=np.linspace(0.6, 1.4, 21))

    # define parameters for bootstrap
    sigma0 = 0.5
    params_dict = {'beta=-1': LogSvParams(sigma0=sigma0, theta=sigma0, kappa1=5.0, kappa2=5.0, beta=-1, volvol=1.0),
                   'beta=0': LogSvParams(sigma0=sigma0, theta=sigma0, kappa1=5.0, kappa2=5.0, beta=0.0, volvol=1.4),
                   'beta=1': LogSvParams(sigma0=sigma0, theta=sigma0, kappa1=5.0, kappa2=5.0, beta=1.0, volvol=1.0)}

    params_dict = {
        'volvol=1.0': LogSvParams(sigma0=sigma0, theta=sigma0, kappa1=2.21, kappa2=2.18, beta=0.0, volvol=1.0),
        'volvol=2.0': LogSvParams(sigma0=sigma0 - 0.005, theta=sigma0 - 0.005, kappa1=2.21, kappa2=2.18, beta=0.0,
                                  volvol=2.0),
        'volvol=3.0': LogSvParams(sigma0=sigma0 - 0.01, theta=sigma0 - 0.01, kappa1=2.21, kappa2=2.18, beta=0.0,
                                  volvol=3.0)}

    # get slice for illustration
    option_slice = option_chain.get_slice(id='2w')
    logsv_pricer.plot_model_slices_in_params(option_slice=option_slice,
                                             params_dict=params_dict)


class UnitTests(Enum):
    PLOT_SKEWS = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.PLOT_SKEWS:
        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(3, 1, figsize=(10, 7))
        plot_skews()

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_SKEWS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
figures for paper
https://www.worldscientific.com/doi/10.1142/S0219024924500031
Log-Normal Stochastic Volatility Model With Quadratic Drift
"""
import string
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from typing import Tuple
from numba.typed import List
from enum import Enum

# chain
from option_chain_analytics import OptionsDataDFs, create_chain_from_from_options_dfs
from option_chain_analytics.ts_loaders import ts_data_loader_wrapper

# analytics
from stochvolmodels import OptionChain, LogSvParams, LogSVPricer, VariableType, ExpansionOrder
from stochvolmodels.pricers.logsv.vol_moments_ode import compute_analytic_qvar
from stochvolmodels.data.fetch_option_chain import generate_vol_chain_np
import stochvolmodels.data.test_option_chain as chains
from stochvolmodels.utils.funcs import set_seed, compute_histogram_data
import stochvolmodels.utils.plots as plot

# implementations for paper
import stochvolmodels.my_papers.logsv_model_wtih_quadratic_drift as mvq
import stochvolmodels.my_papers.logsv_model_wtih_quadratic_drift.steady_state_pdf as ssp
import stochvolmodels.my_papers.logsv_model_wtih_quadratic_drift.ode_sol_in_time as osi
from stochvolmodels.my_papers.logsv_model_wtih_quadratic_drift.model_fit_to_options_timeseries import report_calibration_timeseries


def plot_fitted_model(option_chain: OptionChain,
                      params: LogSvParams,
                      figsize: Tuple[float, float] = (18, 7),
                      fontsize: int = 14
                      ) -> Tuple[plt.Figure, plt.Figure]:
    logsv_pricer = LogSVPricer()
    vol_scaler = logsv_pricer.set_vol_scaler(option_chain=option_chain)

    kwargs = dict(fontsize=fontsize, xvar_format='{:,.0f}')
    fig1 = logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain, params=params, vol_scaler=vol_scaler,
                                                    figsize=figsize,
                                                    **kwargs)
    # fig2 = logsv_pricer.plot_model_ivols_vs_mc(option_chain=option_chain, params=params, nb_path=400000, vol_scaler=vol_scaler,
    #                                           figsize=figsize,
    #                                           **kwargs)

    fig2 = logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=option_chain, params=params, nb_path=400000,
                                                              vol_scaler=vol_scaler,
                                                              figsize=figsize,
                                                              **kwargs)
    return fig1, fig2


def plot_qvar_figure(params: LogSvParams, fontsize: int = 14) -> plt.Figure:

    logsv_pricer = LogSVPricer()

    # ttms = {'1m': 1.0/12.0, '6m': 0.5}
    ttms = {'1w': 7.0/365.0,  '2w': 14.0/365.0, '1m': 1.0/12.0}

    option_chain = chains.get_qv_options_test_chain_data()
    option_chain = OptionChain.get_slices_as_chain(option_chain, ids=list(ttms.keys()))

    forwards = np.array([compute_analytic_qvar(params=params, ttm=ttm, n_terms=4) for ttm in ttms.values()])
    print(f"QV forwards = {forwards}")

    option_chain.forwards = forwards  # replace forwards to imply BSM vols
    option_chain.strikes_ttms = List(forward*strikes_ttm for forward, strikes_ttm in zip(option_chain.forwards, option_chain.strikes_ttms))

    kwargs = dict(fontsize=fontsize)
    fig = logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=option_chain,
                                                             params=params,
                                                             is_plot_vols=True,
                                                             variable_type=VariableType.Q_VAR,
                                                             figsize=(18, 7),
                                                             nb_path=200000,
                                                             **kwargs)
    return fig


def plot_var_pdfs(params: LogSvParams,
                  ttm: float = 1.0,
                  axs: List[plt.Subplot] = None,
                  n: int = 200,
                  vol_scaler: float = None,
                  nb_path: int = 400000,
                  fontsize: int = 14
                  ) -> None:

    logsv_pricer = LogSVPricer()

    # run mc
    x0, sigma0, qvar0 = logsv_pricer.simulate_terminal_values(ttm=ttm, params=params, nb_path=nb_path)

    # normalise quadratic variance
    qvar0 = qvar0 / ttm

    var_datas = {(r'Log-return $X_{\tau}$', VariableType.LOG_RETURN): x0,
                 (r'Quadratic Variance $\frac{I_{\tau}}{\tau}$', VariableType.Q_VAR): qvar0,
                 (r'Volatility $\sigma_{\tau}$', VariableType.SIGMA): sigma0}
    # var_datas = {(r'Log-return $X_{\tau}$', VariableType.LOG_RETURN): x0}
    # var_datas = {(r'Quadratic Variance $\frac{I_{\tau}}{\tau}$', VariableType.Q_VAR): qvar0}
    # var_datas = {('$\sigma_{0}$', VariableType.SIGMA): sigma0}

    if axs is None:
        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 3, figsize=(18, 7), tight_layout=True)

    for idx, (key, mc_data) in enumerate(var_datas.items()):
        variable_type: VariableType = key[1]
        space_grid = params.get_variable_space_grid(variable_type=variable_type, ttm=ttm, n=n, n_stdevs=4.5)
        xpdf1 = logsv_pricer.logsv_pdfs(params=params, ttm=ttm, space_grid=space_grid, variable_type=variable_type,
                                        expansion_order=ExpansionOrder.FIRST,
                                        vol_scaler=vol_scaler,
                                        is_stiff_solver=True)
        xpdf1 = pd.Series(xpdf1, index=space_grid, name='1st order Expansion')
        xpdf2 = logsv_pricer.logsv_pdfs(params=params, ttm=ttm, space_grid=space_grid, variable_type=variable_type,
                                        expansion_order=ExpansionOrder.SECOND,
                                        vol_scaler=vol_scaler,
                                        is_stiff_solver=True)
        xpdf2 = pd.Series(xpdf2, index=space_grid, name='2nd order Expansion')
        xdfs = pd.concat([xpdf1, xpdf2], axis=1)

        mc = compute_histogram_data(data=mc_data, x_grid=space_grid, name='MC')

        df = pd.concat([mc, xdfs], axis=1)
        print(key[0])
        print(df.sum(axis=0))

        ax = axs[idx]
        colors = ['lightblue', 'green', 'brown']
        #sns.lineplot(data=df, dashes=False, palette=colors, fontsize=fontsize, ax=ax)
        qis.plot_line(df=df, dashes=False, colors=colors, fontsize=fontsize,
                      yvar_format='{:,.3f}',
                      legend_loc='upper right',
                      ax=ax)

        ax.fill_between(df.index, np.zeros_like(mc.to_numpy()), mc.to_numpy(),
                        facecolor='lightblue', step='mid', alpha=0.8, lw=1.0)

        ax.set_title(f"({string.ascii_uppercase[idx]}) {key[0]}", color='darkblue')
        ax.set_ylim((0.0, None))
        if variable_type in [VariableType.Q_VAR, VariableType.SIGMA]:
            ax.set_xlim((0.0, None))
        ax.set_xlabel(key[0], fontsize=fontsize)


class UnitTests(Enum):
    FIGURE1_STEADY_STATE = 1
    FIGURE2_VOL_MOMENTS = 2
    FIGURE3_QVAR_EXP = 3
    FIGURE4_FIRST_ORDER = 4
    FIGURE5_SECOND_ORDER = 5
    FIGURE6_JOINT_PDF = 6
    FIGURE7_BTC_CALIBRATIONS = 7
    FIGURE8_9_FITTED_MODEL = 89
    FIGURE10_QVAR = 10


def run_unit_test(unit_test: UnitTests):

    local_path = f"C://Users//artur//OneDrive//My Papers//Working Papers//LogNormal Stochastic Volatility. London. Oct 2013//final_figures//"

    params = LogSvParams(sigma0=0.4083, theta=0.3789, kappa1=2.21, kappa2=2.18, beta=0.5010, volvol=3.0633)
    params.print_vol_moments_stability(n_terms=4)

    ticker = 'BTC'
    value_time = pd.Timestamp('2023-06-30 10:00:00+00:00')

    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker))
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)

    option_chain = generate_vol_chain_np(chain=chain,
                                         value_time=value_time,
                                         days_map={'1w': 7, '2w': 14, '1m': 21},
                                         delta_bounds=(-0.1, 0.1),
                                         is_filtered=True)

    # option_chain.print()

    if unit_test == UnitTests.FIGURE1_STEADY_STATE:

        VOLVOL = 1.5

        SS_PDF_PARAMS = {
            '$(\kappa_{1}=4, \kappa_{2}=0)$': LogSvParams(theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
            '$(\kappa_{1}=4, \kappa_{2}=4)$': LogSvParams(theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
            '$(\kappa_{1}=4, \kappa_{2}=8)$': LogSvParams(theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}

        SS_PDF_PARAMS1 = {
            '$(\kappa_{1}=1)$': LogSvParams(theta=1.0, kappa1=1.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
            '$(\kappa_{1}=4)$': LogSvParams(theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
            '$(\kappa_{1}=8)$': LogSvParams(theta=1.0, kappa1=8.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 3, figsize=(18, 6), tight_layout=True)
            ssp.plot_steady_state(params_dict=SS_PDF_PARAMS,
                                  title='(A) Steady state distribution of the volatility',
                                  ax=axs[0])
            ssp.plot_vol_skew(params_dict=SS_PDF_PARAMS1,
                              title=f'(B) Skeweness of volatility as function of $\kappa_{2}$',
                              ax=axs[1])
            ssp.plot_ss_kurtosis(params_dict=SS_PDF_PARAMS1,
                                 title=f'(C) Excess kurtosis of log-returns as function of $\kappa_{2}$',
                                 ax=axs[2])

            plot.save_fig(fig=fig, local_path=local_path, file_name='figure1_steady_state')
            plot.save_fig(fig=fig, local_path=local_path, file_name='figure1_steady_state', extension='eps', dpi=600)

    elif unit_test == UnitTests.FIGURE2_VOL_MOMENTS:

        set_seed(37)

        params = LogSvParams(sigma0=1.5, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=1.0)

        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 2, figsize=(18, 6), tight_layout=True)
        mvq.plot_vol_moments_vs_mc(params=params,
                                   n_terms=4, n_terms_to_display=4,
                                   title='(A) Volatility moments with $k^{*}=4$',
                                   ax=ax[0])
        mvq.plot_vol_moments_vs_mc(params=params,
                                   n_terms=8, n_terms_to_display=4,
                                   title='(B) Volatility moments with $k^{*}=8$',
                                   ax=ax[1])

        plot.save_fig(fig=fig, local_path=local_path, file_name='figure2_vol_moments')
        plot.save_fig(fig=fig, local_path=local_path, file_name='figure2_vol_moments', extension='eps', dpi=600)

    elif unit_test == UnitTests.FIGURE3_QVAR_EXP:

        set_seed(37)

        VOLVOL = 1.5
        SIGMA0P = 1.5

        TEST_PARAMS = {
            '$(\kappa_{1}=4, \kappa_{2}=0), \sigma_{0}=1.5$': LogSvParams(sigma0=SIGMA0P, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
            '$(\kappa_{1}=4, \kappa_{2}=4), \sigma_{0}=1.5$': LogSvParams(sigma0=SIGMA0P, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
            '$(\kappa_{1}=4, \kappa_{2}=8), \sigma_{0}=1.5$': LogSvParams(sigma0=SIGMA0P, theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}

        TEST_PARAMS2 = {
            '$(\kappa_{1}=4, \kappa_{2}=0), \sigma_{0}=0.5$': LogSvParams(sigma0=0.5, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
            '$(\kappa_{1}=4, \kappa_{2}=4), \sigma_{0}=0.5$': LogSvParams(sigma0=0.5, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
            '$(\kappa_{1}=4, \kappa_{2}=8), \sigma_{0}=0.5$': LogSvParams(sigma0=0.5, theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}

        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(18, 6), tight_layout=True)
            mvq.plot_qvar_vs_mc(params=(TEST_PARAMS | TEST_PARAMS2), ttm=2.0, is_vol=False,
                            title=r'Expected quadratic variance at time $\tau$', n_terms=4, ax=ax)

        plot.save_fig(fig=fig, local_path=local_path, file_name='figure3_qvar_exp')
        plot.save_fig(fig=fig, local_path=local_path, file_name='figure3_qvar_exp', extension='eps', dpi=600)

    if unit_test == UnitTests.FIGURE4_FIRST_ORDER:
        params = LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8606, kappa2=4.7938, beta=0.1985, volvol=2.3690)
        fig = osi.plot_ode_solutions(params=params, ttm=1.0, expansion_order=ExpansionOrder.FIRST,
                                     is_spot_measure=True)
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure4_first_order")
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure4_first_order", extension='eps', dpi=600)

    elif unit_test == UnitTests.FIGURE5_SECOND_ORDER:
        params = LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8606, kappa2=4.7938, beta=0.1985, volvol=2.3690)
        fig = osi.plot_ode_solutions(params=params, ttm=1.0, expansion_order=ExpansionOrder.SECOND,
                                     is_spot_measure=True)
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure5_second_order")
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure5_second_order", extension='eps', dpi=600)

    elif unit_test == UnitTests.FIGURE6_JOINT_PDF:
        set_seed(37)

        params = LogSvParams(sigma0=0.4083, theta=0.3789, kappa1=2.21, kappa2=2.18, beta=0.5010, volvol=0.6 * 3.0633)
        params.print_vol_moments_stability(n_terms=4)

        logsv_pricer = LogSVPricer()
        vol_scaler = logsv_pricer.set_vol_scaler(option_chain=option_chain)

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 3, figsize=(18, 6), tight_layout=True)
        plot_var_pdfs(params=params, ttm=1.0/12.0, vol_scaler=vol_scaler, axs=axs)
        plot.set_subplot_border(fig=fig, n_ax_rows=1, n_ax_col=3)

        plot.save_fig(fig=fig, local_path=local_path, file_name="figure6_joint_pdf")
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure6_joint_pdf", extension='eps', dpi=600)

    elif unit_test == UnitTests.FIGURE7_BTC_CALIBRATIONS:
        df = qis.load_df_from_excel(file_name='btc_calibration_w_fri_20231209_1504', local_path=f"C://Users//artur//OneDrive//analytics//resources//")
        # df = qis.load_df_from_excel(file_name='eth_calibration_w_fri_20231210_1124', local_path=f"C://Users//artur//OneDrive//analytics//resources//")
        fig = report_calibration_timeseries(df=df)
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure7_btc_calibrations")
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure7_btc_calibrations", extension='eps', dpi=600)

    elif unit_test == UnitTests.FIGURE8_9_FITTED_MODEL:
        fig1, fig2 = plot_fitted_model(option_chain=option_chain,
                                       params=params)

        plot.save_fig(fig=fig1, local_path=local_path, file_name="figure8_btc_fit")
        plot.save_fig(fig=fig1, local_path=local_path, file_name="figure8_btc_fit", extension='eps', dpi=600)
        plot.save_fig(fig=fig2, local_path=local_path, file_name="figure9_btc_mc_comp")
        plot.save_fig(fig=fig2, local_path=local_path, file_name="figure9_btc_mc_comp", extension='eps', dpi=600)

    elif unit_test == UnitTests.FIGURE10_QVAR:

        set_seed(20)
        fig = plot_qvar_figure(params=params)
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure10_qvar")
        plot.save_fig(fig=fig, local_path=local_path, file_name="figure10_qvar", extension='eps', dpi=600)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.FIGURE1_STEADY_STATE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
figures 6 and 7 in the logsv_model_wtih_quadratic_drift
plot moments of vol and quadratic variance in time
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import seaborn as sns
from typing import Dict
from enum import Enum

import stochvolmodels.pricers.logsv.vol_moments_ode as vmo
import stochvolmodels.utils.plots as plot
from stochvolmodels.utils.funcs import set_seed
from stochvolmodels.pricers.logsv_pricer import LogSVPricer, LogSvParams

VOLVOL = 1.5
SIGMA0P = 1.5


TEST_PARAMS = {'$(\kappa_{1}=4, \kappa_{2}=0), \sigma_{0}=1.5$': LogSvParams(sigma0=SIGMA0P, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
               '$(\kappa_{1}=4, \kappa_{2}=4), \sigma_{0}=1.5$': LogSvParams(sigma0=SIGMA0P, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
               '$(\kappa_{1}=4, \kappa_{2}=8), \sigma_{0}=1.5$': LogSvParams(sigma0=SIGMA0P, theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}

TEST_PARAMS2 = {'$(\kappa_{1}=4, \kappa_{2}=0), \sigma_{0}=0.5$': LogSvParams(sigma0=0.5, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
                '$(\kappa_{1}=4, \kappa_{2}=4), \sigma_{0}=0.5$': LogSvParams(sigma0=0.5, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
                '$(\kappa_{1}=4, \kappa_{2}=8), \sigma_{0}=0.5$': LogSvParams(sigma0=0.5, theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}


def plot_vol_moments_vs_mc(params: LogSvParams = LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
                           ttm: float = 1.5,
                           n_terms: int = 4,
                           n_terms_to_display: int = 4,
                           nb_path: int = 100000,
                           grid_step: int = 15,
                           title: str = 'Volatility moments',
                           ax: plt.Subplot = None
                           ) -> None:
    """
    plot mc comparison with old_analytics for vol moments
    """
    logsv_pricer = LogSVPricer()
    params.assert_vol_moments_stability(n_terms=n_terms)
    sigma_t, grid_t = logsv_pricer.simulate_vol_paths(ttm=ttm, params=params, nb_path=nb_path, year_days=360)

    palette = plot.get_n_sns_colors(n=n_terms_to_display)
    if ax is None:
        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)

    for n in np.arange(n_terms_to_display):
        if n > 0:
            m_n = np.power(sigma_t-params.theta, n+1)
        else:
            m_n = sigma_t - params.theta
        mc_mean, mc_std = np.mean(m_n, axis=1), np.std(m_n, axis=1) / np.sqrt(nb_path)

        ax.errorbar(x=grid_t[::grid_step], y=mc_mean[::grid_step], yerr=1.96*mc_std[::grid_step], fmt='o',
                    color=palette[n],
                    markersize=3, capsize=3)

    analytic_vol_moments = vmo.compute_vol_moments_t(params=params, ttm=grid_t, n_terms=n_terms)
    if n_terms > n_terms_to_display:
        analytic_vol_moments = analytic_vol_moments[:, :n_terms_to_display]
    analytic_vol_moments = pd.DataFrame(analytic_vol_moments, index=grid_t, columns=[f"Analytic m{n + 1}" for n in range(n_terms_to_display)])

    sns.lineplot(data=analytic_vol_moments, dashes=True, palette=palette, ax=ax)
    if title is not None:
        ax.set_title(title, fontsize=12, color='darkblue')
    ax.set_xlabel(r'$\tau$')
    ax.set_xlim((0.0, None))


def plot_qvar_vs_mc(params: Dict[str, LogSvParams] = TEST_PARAMS,
                    ttm: float = 1.5,
                    n_terms: int = 4,
                    nb_path: int = 100000,
                    grid_step: int = 15,
                    is_vol: bool = True,
                    title: str = 'Expected values',
                    ax:  plt.Subplot = None
                    ) -> None:
    """
    plot mc comparison for old_analytics with expected vol / qvar curves in t
    """
    logsv_pricer = LogSVPricer()

    analytic_vol_moments = []
    colors = plot.get_n_sns_colors(n=len(params.keys())//2)
    palette = colors + colors

    for idx, (key, params_) in enumerate(params.items()):
        params_.assert_vol_moments_stability(n_terms=n_terms)
        sigma_t, grid_t = logsv_pricer.simulate_vol_paths(ttm=ttm, params=params_, nb_path=nb_path, year_days=360)

        if is_vol:
            mc_mean = np.mean(sigma_t, axis=1)
            mc_std = 2.0*1.96 * np.std(sigma_t, axis=1) / np.sqrt(nb_path)
        else:
            q_var = pd.DataFrame(np.square(sigma_t)).expanding(axis=0).mean().to_numpy()
            mc_mean = np.mean(q_var, axis=1)
            mc_std = 2.0*1.96*np.std(q_var, axis=1) / np.sqrt(nb_path)

        ax.errorbar(x=grid_t[::grid_step], y=mc_mean[::grid_step], yerr=mc_std[::grid_step], fmt='o', color=palette[idx],
                    markersize=3, capsize=3)

        if is_vol:
            v_t = vmo.compute_expected_vol_t(params=params_, t=grid_t, n_terms=n_terms)
        else:
            v_t = vmo.compute_sqrt_qvar_t(params=params_, t=grid_t, n_terms=n_terms)
            v_t = np.square(v_t)
        analytic_vol_moments.append(pd.Series(v_t, index=grid_t, name=f"Analytic {key}"))

    analytic_vol_moments = pd.concat(analytic_vol_moments, axis=1)
    sns.lineplot(data=analytic_vol_moments, dashes=True, palette=palette, ax=ax)
    if title is not None:
        ax.set_title(title, fontsize=12, color='darkblue')
    ax.set_xlabel(r'$\tau$')
    ax.set_xlim((0.0, None))


def plot_expected_vol_qvar(params: Dict[str, LogSvParams],
                           ttm: float = 1.5,
                           is_vol: bool = True,
                           n_terms: int = 4,
                           title: str = 'Expected values',
                           ax:  plt.Subplot = None
                           ) -> None:
    """
    plot expected vol / qvar curves in t
    """
    evs = []
    t = np.linspace(0.0, ttm, 50)
    for key, param in params.items():
        param.assert_vol_moments_stability(n_terms=n_terms)
        if is_vol:
            v_t = vmo.compute_expected_vol_t(params=param, t=t, n_terms=n_terms)
        else:
            v_t = vmo.compute_sqrt_qvar_t(params=param, t=t, n_terms=n_terms)
        evs.append(pd.Series(v_t, index=t, name=key))
    evs = pd.concat(evs, axis=1)

    colors = plot.get_n_sns_colors(n=len(params.keys())//2)
    palette = colors + colors
    sns.lineplot(data=evs, dashes=True, palette=palette, ax=ax)
    yvar_format = '{:.0%}'
    ax.yaxis.set_major_formatter(mticker.FuncFormatter(lambda z, _: yvar_format.format(z)))

    if title is not None:
        ax.set_title(title, fontsize=12, color='darkblue')
    ax.set_xlabel(r"$\tau$", fontsize=12)
    ax.set_xlim((0.0, None))


class UnitTests(Enum):
    VOL_MOMENTS_VS_MC = 1
    QVAR_VS_MC = 2
    JOINT_FIGURE = 3
    JOINT_VOL_MOMENTS_VS_MC_FIGURE = 4
    SINGLE_QVAR_FIGURE = 5


def run_unit_test(unit_test: UnitTests):

    set_seed(37) #33
    ttm = 2.0

    if unit_test == UnitTests.VOL_MOMENTS_VS_MC:
        params = LogSvParams(sigma0=1.5, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=1.0)

        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(18, 7), tight_layout=True)
        plot_vol_moments_vs_mc(params=params,
                               n_terms=4, n_terms_to_display=4,
                               ax=ax)

    elif unit_test == UnitTests.QVAR_VS_MC:
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(1, 2, figsize=(18, 10), tight_layout=True)
            plot_qvar_vs_mc(params=(TEST_PARAMS | TEST_PARAMS2), ttm=ttm, is_vol=True,
                                   title='(A) Expected volatility at time T', n_terms=4, ax=axs[0])
            plot_qvar_vs_mc(params=(TEST_PARAMS | TEST_PARAMS2), ttm=ttm, is_vol=False,
                                   title='(B) Expected quadratic variance at time T', n_terms=4, ax=axs[1])

    elif unit_test == UnitTests.JOINT_FIGURE:
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(1, 2, figsize=(18, 10), tight_layout=True)
            plot_expected_vol_qvar(params=(TEST_PARAMS | TEST_PARAMS2), ttm=ttm, is_vol=True,
                                   title='(A) Expected volatility at time T', n_terms=4, ax=axs[0])
            plot_expected_vol_qvar(params=(TEST_PARAMS | TEST_PARAMS2), ttm=ttm, is_vol=False,
                                   title='(B) Expected quadratic variance at time T', n_terms=4, ax=axs[1])

        is_save = True
        if is_save:
            plot.save_fig(fig=fig, local_path='../../../docs/figures//', file_name='vol_qvar_exp')

    elif unit_test == UnitTests.JOINT_VOL_MOMENTS_VS_MC_FIGURE:
        params = LogSvParams(sigma0=1.5, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=1.0)

        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 2, figsize=(18, 6), tight_layout=True)
        plot_vol_moments_vs_mc(params=params,
                               n_terms=4, n_terms_to_display=4,
                               title='(A) Volatility moments with $k^{*}=4$',
                               ax=ax[0])
        plot_vol_moments_vs_mc(params=params,
                               n_terms=8, n_terms_to_display=4,
                               title='(B) Volatility moments with $k^{*}=8$',
                               ax=ax[1])

        is_save = False
        if is_save:
            plot.save_fig(fig=fig, local_path='../../../docs/figures//', file_name='vol_moments')

    elif unit_test == UnitTests.SINGLE_QVAR_FIGURE:
        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(18, 6), tight_layout=True)
            plot_qvar_vs_mc(params=(TEST_PARAMS | TEST_PARAMS2), ttm=ttm, is_vol=False,
                            title=r'Expected quadratic variance at time $\tau$', n_terms=4, ax=ax)

        is_save = False
        if is_save:
            plot.save_fig(fig=fig, local_path='../../../docs/figures//', file_name='qvar_exp')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.JOINT_VOL_MOMENTS_VS_MC_FIGURE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
fetch vol data either using historical ohlc vol or VIX and the likes
"""
import numpy as np
import pandas as pd
import qis
import yfinance as yf
from typing import Optional, Tuple
from qis import OhlcEstimatorType


def fetch_ohlc_vol(ticker: str = 'SPY',
                   af: float = 260,
                   timeperiod: Optional[qis.TimePeriod] = qis.TimePeriod('31Dec1999', None),
                   ohlc_estimator_type: OhlcEstimatorType = OhlcEstimatorType.ROGERS_SATCHELL
                   ) -> Tuple[pd.Series, pd.Series]:
    if ticker in ['VIX', 'MOVE', 'OVX']:  # use implied indices
        ohlc_data = yf.download(tickers=f"^{ticker}", start=None, end=None, ignore_tz=True)
        ohlc_data.index = ohlc_data.index.tz_localize('UTC').tz_convert('UTC')
        vol = ohlc_data['Close'] / 100.0

        if ticker == 'VIX':
            spot_ticker = '^GSPC'  # s&p 500 index
        elif ticker == 'MOVE':
            spot_ticker = '^TNX'   # 10y rate
        elif ticker == 'OVX':
            spot_ticker = 'USO'  # oil fund
        else:
            raise NotImplementedError

        prices = yf.download(tickers=spot_ticker, start=None, end=None, ignore_tz=True)['Adj Close']
        prices.index = prices.index.tz_localize('UTC').tz_convert('UTC')

        if ticker == 'MOVE':
            returns = prices.diff(1)
        else:
            # returns = np.log(prices).diff(1)
            returns = prices.pct_change()

    elif ticker in ['BTC', 'ETH']:  # use implied atm vols from internal data
        df = qis.load_df_from_csv(file_name=f"{ticker}_atm_vols",
                                  local_path="C://Users//Artur//OneDrive//analytics//resources//")
        prices = df.iloc[:, 0]
        vol = df.iloc[:, -1]
        returns = prices.pct_change()

    else:  # use historical vol
        data = yf.download(tickers=ticker, start=None, end=None, ignore_tz=True)
        data.index = data.index.tz_localize('UTC').tz_convert('UTC')
        ohlc_data = data[['Open', 'High', 'Low', 'Close']].rename({'Open': 'open', 'High': 'high', 'Low': 'low', 'Close': 'close'}, axis=1)
        var = qis.estimate_ohlc_var(ohlc_data=ohlc_data, ohlc_estimator_type=ohlc_estimator_type)
        vol = np.sqrt(af*var)

        returns = np.log(data['Adj Close']).diff(1)

    vol = vol.replace([0.0, np.inf, -np.inf], np.nan).dropna()  # drop outliers

    if timeperiod is not None:
        vol = timeperiod.locate(vol)
        returns = timeperiod.locate(returns)

    vol = vol.rename(ticker)
    returns = returns.rename(ticker)
    return vol, returns



"""
figure 4 in the logsv_model_wtih_quadratic_drift
plot volatility drift
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import seaborn as sns
from typing import Dict, List
from enum import Enum

import stochvolmodels.utils.plots as plot
from stochvolmodels.pricers.logsv_pricer import LogSvParams

VOLVOL = 1.75


DRIFT_PARAMS = {'$(\kappa_{1}=4, \kappa_{2}=0)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
                '$(\kappa_{1}=4, \kappa_{2}=4)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
                '$(\kappa_{1}=4, \kappa_{2}=8)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}


def plot_drift(params: Dict[str, LogSvParams] = DRIFT_PARAMS,
               axs: List[plt.Subplot] = None
               ) -> None:
    x = np.linspace(0.0, 2.0, 200)

    drifts = []
    drifts_delta = []
    for key, param in params.items():
        drift1 = param.kappa1*(param.theta - x)
        drift2 = param.kappa1 * param.theta - (param.kappa1 - param.kappa2 * param.theta) * x - param.kappa2 * x * x
        drifts.append(pd.Series(drift2, index=x, name=key))
        drifts_delta.append(pd.Series(drift2-drift1, index=x, name=key))
    drifts = pd.concat(drifts, axis=1) / 260.0
    drifts_delta = pd.concat(drifts_delta, axis=1) / 260.0

    dfs = {'(A) Volatility drift per day as function of $\sigma_{t}$': drifts,
           '(B) Volatility drift relative to the linear drift': drifts_delta}

    for idx, (key, df) in enumerate(dfs.items()):
        ax = axs[idx]
        sns.lineplot(data=df, dashes=False,  ax=ax)
        yvar_format = '{:.2f}'
        ax.yaxis.set_major_formatter(mticker.FuncFormatter(lambda z, _: yvar_format.format(z)))
        ax.set_title(key, fontsize=12, color='darkblue')
        ax.set_xlabel('$\sigma_{t}$', fontsize=12)
        ax.set_xlim((0.0, None))
    plot.align_y_limits_axs(axs=axs)


class UnitTests(Enum):
    PLOT_DRIFT = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.PLOT_DRIFT:
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(1, 2, figsize=(18, 6), tight_layout=True)
        plot_drift(axs=axs)

        is_save = False
        if is_save:
            plot.save_fig(fig=fig,
                          local_path='../../../docs/figures//',
                          file_name='vol_drift')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_DRIFT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




import numpy as np
import pandas as pd
import qis as qis
import matplotlib.pyplot as plt


def plot_kernels(kappa: float = -2.0):
    x = np.linspace(-0.5, 1.0, 1000)
    exp_k = pd.Series(np.exp(x*kappa), index=x, name='Exp')
    #  linear_k = pd.Series((np.power(1.0+x, 1-kappa)-1.0)/(1-kappa), index=x, name='Linear')
    linear_k = pd.Series(1.0 + x*kappa + 0.5*np.square(x*kappa) + (1.0/6.0)*np.square(x*kappa)*x*kappa, index=x, name='Linear')
    df = pd.concat([exp_k, linear_k], axis=1)
    qis.plot_line(df=df)


plot_kernels()

plt.show()




"""
figures 11, 12, 13 and data for table 1
model calibration to implied volatilities of different assets
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from enum import Enum

from stochvolmodels.utils.config import VariableType
from stochvolmodels.data.option_chain import OptionChain
from stochvolmodels.pricers.logsv_pricer import LogSVPricer, LogsvModelCalibrationType, ConstraintsType, LogSvParams
from stochvolmodels.pricers.logsv.vol_moments_ode import compute_analytic_qvar
from stochvolmodels.utils.funcs import set_seed
import stochvolmodels.utils.plots as plot
import stochvolmodels.data.test_option_chain as chains


class Assets(str, Enum):
    BTC = 'Bitcoin'
    VIX = 'Vix'
    GLD = 'Gold'
    SQQQ = '-3x Nasdaq'
    SPY = 'S&P500'


CALIBRATED_PARAMS = {
    Assets.VIX: LogSvParams(sigma0=0.9767, theta=0.5641, kappa1=4.9067, kappa2=8.6985, beta=2.3425, volvol=1.0163),
    Assets.SQQQ: LogSvParams(sigma0=0.9114, theta=0.9390, kappa1=4.9544, kappa2=5.2762, beta=1.3215, volvol=0.9964),
    Assets.BTC: LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8609, kappa2=4.7940, beta=0.1988, volvol=2.3694),
    Assets.GLD: LogSvParams(sigma0=0.1505, theta=0.1994, kappa1=2.2062, kappa2=11.0630, beta=0.1547, volvol=2.8011),
    Assets.SPY: LogSvParams(sigma0=0.2270, theta=0.2616, kappa1=4.9325, kappa2=18.8550, beta=-1.8123, volvol=0.9832),
}


def get_asset_chain_data(asset: Assets = Assets.BTC) -> OptionChain:
    match asset:
        case Assets.BTC:
            option_chain = chains.get_btc_test_chain_data()
        case Assets.VIX:
            option_chain = chains.get_vix_test_chain_data()
        case Assets.GLD:
            option_chain = chains.get_gld_test_chain_data()
        case Assets.SQQQ:
            option_chain = chains.get_sqqq_test_chain_data()
        case Assets.SPY:
            option_chain = chains.get_spy_test_chain_data()
        case _:
            raise NotImplementedError(f"not implemented {asset}")
    return option_chain


def calibrate_logsv_model(asset: Assets = Assets.BTC,
                          model_calibration_type: LogsvModelCalibrationType = LogsvModelCalibrationType.PARAMS5
                          ):
    match asset:
        case Assets.BTC:
            params0 = LogSvParams(sigma0=0.84, theta=1.04, kappa1=5.0, kappa2=None, beta=0.15, volvol=1.85)
            constraints_type = ConstraintsType.INVERSE_MARTINGALE
        case Assets.VIX:
            params0 = LogSvParams(sigma0=0.8, theta=0.6, kappa1=5.0, kappa2=None, beta=2.0, volvol=1.0)
            constraints_type = ConstraintsType.MMA_MARTINGALE_MOMENT4
        case Assets.GLD:
            params0 = LogSvParams(sigma0=0.1530, theta=0.1960, kappa1=2.2068, kappa2=11.2584, beta=0.1580, volvol=2.8022)
            constraints_type = ConstraintsType.UNCONSTRAINT
        case Assets.SQQQ:
            params0 = LogSvParams(sigma0=1.0, theta=1.0, kappa1=5.0, kappa2=None, beta=1.0, volvol=1.0)
            constraints_type = ConstraintsType.MMA_MARTINGALE_MOMENT4
        case Assets.SPY:
            params0 = LogSvParams(sigma0=0.2, theta=0.2, kappa1=5.0, kappa2=None, beta=-1.0, volvol=1.0)
            constraints_type = ConstraintsType.MMA_MARTINGALE_MOMENT4
        case _:
            raise NotImplementedError(f"not implemented {asset}")

    option_chain = get_asset_chain_data(asset=asset)
    logsv_pricer = LogSVPricer()
    fit_params = logsv_pricer.calibrate_model_params_to_chain(option_chain=option_chain,
                                                              params0=params0,
                                                              model_calibration_type=model_calibration_type,
                                                              constraints_type=constraints_type)
    fit_params.print_vol_moments_stability()
    print(fit_params)
    fig = logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain,
                                                   params=fit_params,
                                                   headers=('(A)', '(B)', '(C)', '(D)'))

    return fig


class UnitTests(Enum):
    CALIBRATION = 1
    MODEL_COMPARISION_WITH_MC = 2
    ALL_PARAMS_TABLE = 3
    PLOT_BTC_COMP_FOR_ARTICLE = 4
    PLOT_QVAR_FIGURE_FOR_ARTICLE = 5


def run_unit_test(unit_test: UnitTests):

    set_seed(24)

    if unit_test == UnitTests.CALIBRATION:
        asset = Assets.BTC
        fig = calibrate_logsv_model(asset=asset)
        plot.save_fig(fig=fig, local_path='../../../docs/figures//', file_name=f"calibration_{asset.value}")

    elif unit_test == UnitTests.MODEL_COMPARISION_WITH_MC:

        asset = Assets.BTC
        option_chain = get_asset_chain_data(asset=asset)

        params = CALIBRATED_PARAMS[asset]
        params.print_vol_moments_stability()

        logsv_pricer = LogSVPricer()
        # logsv_pricer.plot_model_ivols(option_chain=option_chain, params=params)
        # logsv_pricer.plot_model_ivols_vs_mc(option_chain=option_chain, params=params, nb_path=400000)
        uniform_chain_data = OptionChain.to_uniform_strikes(obj=option_chain, num_strikes=31)
        logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=uniform_chain_data, params=params, nb_path=400000)

    elif unit_test == UnitTests.ALL_PARAMS_TABLE:
        datas = {key.value: param.to_dict() for key, param in CALIBRATED_PARAMS.items()}
        df = pd.DataFrame.from_dict(datas)
        print(df)
        df.to_clipboard()

    elif unit_test == UnitTests.PLOT_BTC_COMP_FOR_ARTICLE:

        asset = Assets.BTC
        params = CALIBRATED_PARAMS[asset]
        option_chain = get_asset_chain_data(asset=asset)
        logsv_pricer = LogSVPricer()

        fig1 = logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain,
                                                        params=params,
                                                        headers=('(A)', '(B)', '(C)', '(D)'))

        uniform_chain_data = OptionChain.to_uniform_strikes(obj=option_chain, num_strikes=31)
        fig2 = logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=uniform_chain_data,
                                                                  params=params,
                                                                  nb_path=400000)

        is_save = False
        if is_save:
            plot.save_fig(fig=fig1, local_path='../../../docs/figures//', file_name="btc_fit")
            plot.save_fig(fig=fig2, local_path='../../../docs/figures//', file_name="btc_mc_comp")

    elif unit_test == UnitTests.PLOT_QVAR_FIGURE_FOR_ARTICLE:

        asset = Assets.BTC
        params = CALIBRATED_PARAMS[asset]
        logsv_pricer = LogSVPricer()

        # ttms = {'1m': 1.0/12.0, '6m': 0.5}
        ttms = {'1m': 1.0 / 12.0, '3m': 0.25}
        forwards = np.array([compute_analytic_qvar(params=params, ttm=ttm) for ttm in ttms.values()])
        print(f"QV forwards = {forwards}")

        option_chain = chains.get_qv_options_test_chain_data()
        option_chain = OptionChain.get_slices_as_chain(option_chain, ids=list(ttms.keys()))
        option_chain.forwards = forwards  # replace forwards to imply BSM vols

        set_seed(80)
        fig = logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=option_chain,
                                                                 params=params,
                                                                 is_plot_vols=True,
                                                                 variable_type=VariableType.Q_VAR,
                                                                 figsize=(18, 6),
                                                                 nb_path=100000)
        is_save = False
        if is_save:
            plot.save_fig(fig=fig, local_path='../../../docs/figures//', file_name="model_vs_mc_qvar_logsv")

    else:
        raise NotImplementedError(f"not implemented {unit_test}")

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_QVAR_FIGURE_FOR_ARTICLE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
